# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for libFuzzer engine."""
# pylint: disable=unused-argument

import os
import shutil
import tempfile
import unittest

import mock
import parameterized
import pyfakefs.fake_filesystem_unittest as fake_fs_unittest
import six

from clusterfuzz._internal.bot.fuzzers import engine_common
from clusterfuzz._internal.bot.fuzzers import libfuzzer
from clusterfuzz._internal.bot.fuzzers import strategy_selection
from clusterfuzz._internal.bot.fuzzers import utils as fuzzer_utils
from clusterfuzz._internal.bot.fuzzers.libFuzzer import constants
from clusterfuzz._internal.bot.fuzzers.libFuzzer import engine
from clusterfuzz._internal.build_management import build_manager
from clusterfuzz._internal.fuzzing import strategy
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.platforms import android
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import new_process
from clusterfuzz._internal.system import process_handler
from clusterfuzz._internal.system import shell
from clusterfuzz._internal.tests.test_libs import android_helpers
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils

try:
  from shlex import quote
except ImportError:
  from pipes import quote

TEST_PATH = os.path.abspath(os.path.dirname(__file__))
TEST_DIR = os.path.join(TEST_PATH, 'libfuzzer_test_data')
TEMP_DIR = os.path.join(TEST_PATH, 'temp')
DATA_DIR = os.path.join(TEST_PATH, 'data')
ANDROID_DATA_DIR = os.path.join(DATA_DIR, 'android')

_get_directory_file_count_orig = shell.get_directory_file_count


class PrepareTest(fake_fs_unittest.TestCase):
  """Prepare() tests."""

  def setUp(self):
    """Setup for fake filesystem prepare test."""
    test_helpers.patch_environ(self)
    test_utils.set_up_pyfakefs(self)

    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.engine_common.unpack_seed_corpus_if_needed',
    ])

    self.fs.create_dir('/inputs')
    self.fs.create_file('/path/target')
    self.fs.create_file('/path/blah.dict')
    self.fs.create_file('/path/target_seed_corpus.zip')
    self.fs.create_file(
        '/path/target.options',
        contents=('[libfuzzer]\n'
                  'max_len=31337\n'
                  'timeout=11\n'
                  'dict=blah.dict\n'))

    os.environ['FAIL_RETRIES'] = '1'
    os.environ['FUZZ_INPUTS_DISK'] = '/inputs'

    test_helpers.patch(
        self, ['clusterfuzz._internal.bot.fuzzers.libfuzzer.pick_strategies'])

    self.mock.pick_strategies.return_value = libfuzzer.StrategyInfo(
        fuzzing_strategies=[
            'unknown_1', 'value_profile', 'corpus_subset_20', 'fork_2'
        ],
        arguments=['-arg1'],
        additional_corpus_dirs=['/new_corpus_dir'],
        extra_env={'extra_env': '1'},
        use_dataflow_tracing=False,
        is_mutations_run=True)

  def test_prepare(self):
    """Test prepare."""
    engine_impl = engine.LibFuzzerEngine()
    options = engine_impl.prepare('/corpus_dir', '/path/target', '/path')
    self.assertEqual('/corpus_dir', options.corpus_dir)
    six.assertCountEqual(self, [
        '-max_len=31337', '-timeout=11', '-rss_limit_mb=2560', '-arg1',
        '-dict=/path/blah.dict'
    ], options.arguments)
    self.assertDictEqual({
        'value_profile': 1,
        'corpus_subset': 20,
        'fork': 2
    }, options.strategies)
    six.assertCountEqual(self, ['/new_corpus_dir', '/corpus_dir'],
                         options.fuzz_corpus_dirs)
    self.assertDictEqual({'extra_env': '1'}, options.extra_env)
    self.assertFalse(options.use_dataflow_tracing)
    self.assertTrue(options.is_mutations_run)

    self.mock.unpack_seed_corpus_if_needed.assert_called_with(
        '/path/target', '/corpus_dir')

  def test_prepare_invalid_dict(self):
    """Test prepare with an invalid dict path."""
    with open('/path/target.options', 'w') as f:
      f.write('[libfuzzer]\n'
              'max_len=31337\n'
              'timeout=11\n'
              'dict=not_exist.dict\n')

    engine_impl = engine.LibFuzzerEngine()
    options = engine_impl.prepare('/corpus_dir', '/path/target', '/path')
    six.assertCountEqual(
        self, ['-max_len=31337', '-timeout=11', '-rss_limit_mb=2560', '-arg1'],
        options.arguments)

  def test_prepare_auto_add_dict(self):
    """Test prepare automatically adding dict argument."""
    with open('/path/target.options', 'w') as f:
      f.write('[libfuzzer]\n' 'max_len=31337\n' 'timeout=11\n')
    self.fs.create_file('/path/target.dict')

    engine_impl = engine.LibFuzzerEngine()
    options = engine_impl.prepare('/corpus_dir', '/path/target', '/path')
    six.assertCountEqual(self, [
        '-max_len=31337', '-timeout=11', '-rss_limit_mb=2560', '-arg1',
        '-dict=/path/target.dict'
    ], options.arguments)


class FuzzAdditionalProcessingTimeoutTest(unittest.TestCase):
  """fuzz_additional_processing_timeout tests."""

  def test_no_mutations(self):
    """Test no mutations."""
    engine_impl = engine.LibFuzzerEngine()
    options = engine.LibFuzzerOptions(
        '/corpus',
        arguments=[],
        strategies=[],
        fuzz_corpus_dirs=[],
        extra_env={},
        use_dataflow_tracing=False,
        is_mutations_run=False)
    self.assertEqual(2100.0,
                     engine_impl.fuzz_additional_processing_timeout(options))

  def test_mutations(self):
    """Test with mutations."""
    engine_impl = engine.LibFuzzerEngine()
    options = engine.LibFuzzerOptions(
        '/corpus',
        arguments=[],
        strategies=[],
        fuzz_corpus_dirs=[],
        extra_env={},
        use_dataflow_tracing=False,
        is_mutations_run=True)
    self.assertEqual(2700.0,
                     engine_impl.fuzz_additional_processing_timeout(options))


class PickStrategiesTest(fake_fs_unittest.TestCase):
  """pick_strategies tests."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.engine_common.is_lpm_fuzz_target',
        'random.SystemRandom.randint',
    ])
    self.mock.is_lpm_fuzz_target.return_value = False

    test_utils.set_up_pyfakefs(self)
    self.fs.create_dir('/path/corpus')
    self.fs.create_file('/path/target')

  def test_max_length_strategy_with_override(self):
    """Tests max length strategy with override."""
    strategy_pool = set_strategy_pool([strategy.RANDOM_MAX_LENGTH_STRATEGY])
    strategy_info = libfuzzer.pick_strategies(strategy_pool, '/path/target',
                                              '/path/corpus', ['-max_len=100'])
    six.assertCountEqual(self, [], strategy_info.arguments)

  def test_max_length_strategy_without_override(self):
    """Tests max length strategy without override."""
    self.mock.randint.return_value = 1337
    strategy_pool = set_strategy_pool([strategy.RANDOM_MAX_LENGTH_STRATEGY])
    strategy_info = libfuzzer.pick_strategies(strategy_pool, '/path/target',
                                              '/path/corpus', [])
    six.assertCountEqual(self, ['-max_len=1337'], strategy_info.arguments)


class FuzzTest(fake_fs_unittest.TestCase):
  """Fuzz() tests."""

  def setUp(self):
    """Setup for fake filesystem fuzz test."""
    test_helpers.patch_environ(self)
    test_utils.set_up_pyfakefs(self)

    self.fs.create_dir('/corpus')
    self.fs.create_dir('/fuzz-inputs')
    self.fs.create_dir('/fake')
    self.fs.create_file('/target')
    self.fs.add_real_directory(TEST_DIR)

    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.libFuzzer.engine._is_multistep_merge_supported',
        'clusterfuzz._internal.bot.fuzzers.libfuzzer.LibFuzzerRunner.fuzz',
        'clusterfuzz._internal.bot.fuzzers.libfuzzer.LibFuzzerRunner.merge',
        'os.getpid',
    ])

    os.environ['TEST_TIMEOUT'] = '65'
    os.environ['JOB_NAME'] = 'libfuzzer_asan_job'
    os.environ['FUZZ_INPUTS_DISK'] = '/fuzz-inputs'

    self.mock._is_multistep_merge_supported = True  # pylint: disable=protected-access
    self.mock.getpid.return_value = 9001
    self.maxDiff = None  # pylint: disable=invalid-name

  def test_fuzz(self):
    """Test fuzz."""
    engine_impl = engine.LibFuzzerEngine()
    options = engine.LibFuzzerOptions('/corpus', [
        '-arg=1',
        '-timeout=123',
        '-dict=blah.dict',
        '-max_len=9001',
        '-use_value_profile=1',
    ], [], ['/corpus'], {}, False, False)

    with open(os.path.join(TEST_DIR, 'crash.txt')) as f:
      fuzz_output = f.read()

    def mock_fuzz(*args, **kwargs):  # pylint: disable=unused-argument
      """Mock fuzz."""
      self.fs.create_file('/fuzz-inputs/temp-9001/new/A')
      self.fs.create_file('/fuzz-inputs/temp-9001/new/B')
      return new_process.ProcessResult(
          command='command',
          return_code=0,
          output=fuzz_output,
          time_executed=2.0,
          timed_out=False)

    # Record the merge calls manually as the mock module duplicates the second
    # call and overwrites the first call arguments.
    mock_merge_calls = []

    def mock_merge(*args, **kwargs):  # pylint: disable=unused-argument
      """Mock merge."""
      mock_merge_calls.append(self.mock.merge.mock_calls[-1])
      self.assertTrue(len(mock_merge_calls) <= 2)

      merge_output_file = 'merge_step_%d.txt' % len(mock_merge_calls)
      with open(os.path.join(TEST_DIR, merge_output_file)) as f:
        merge_output = f.read()

      self.fs.create_file('/fuzz-inputs/temp-9001/merge-corpus/A')
      return new_process.ProcessResult(
          command='merge-command',
          return_code=0,
          output=merge_output,
          time_executed=2.0,
          timed_out=False)

    self.mock.fuzz.side_effect = mock_fuzz
    self.mock.merge.side_effect = mock_merge

    result = engine_impl.fuzz('/target', options, '/fake', 3600)
    self.assertEqual(1, len(result.crashes))
    self.assertEqual(fuzz_output, result.logs)

    crash = result.crashes[0]
    self.assertEqual('/fake/crash-1e15825e6f0b2240a5af75d84214adda1b6b5340',
                     crash.input_path)
    self.assertEqual(fuzz_output, crash.stacktrace)
    six.assertCountEqual(self, ['-arg=1', '-timeout=60'], crash.reproduce_args)
    self.assertEqual(2, crash.crash_time)

    self.mock.fuzz.assert_called_with(
        mock.ANY, ['/fuzz-inputs/temp-9001/new', '/corpus'],
        additional_args=[
            '-arg=1',
            '-timeout=123',
            '-dict=blah.dict',
            '-max_len=9001',
            '-use_value_profile=1',
        ],
        artifact_prefix='/fake',
        extra_env={},
        fuzz_timeout=3600)

    self.assertEqual(2, len(mock_merge_calls))

    # Main things to test are:
    # 1) The new corpus directory is used in the second call only.
    # 2) the merge control file is explicitly specified for both calls.
    mock_merge_calls[0].assert_called_with(
        mock.ANY, [
            '/fuzz-inputs/temp-9001/merge-corpus',
            '/corpus',
        ],
        additional_args=[
            '-arg=1',
            '-timeout=123',
            '-merge_control_file=/fuzz-inputs/temp-9001/merge-workdir/MCF',
        ],
        artifact_prefix=None,
        merge_timeout=1800.0,
        tmp_dir='/fuzz-inputs/temp-9001/merge-workdir')

    mock_merge_calls[1].assert_called_with(
        mock.ANY, [
            '/fuzz-inputs/temp-9001/merge-corpus',
            '/corpus',
            '/fuzz-inputs/temp-9001/new',
        ],
        additional_args=[
            '-arg=1',
            '-timeout=123',
            '-merge_control_file=/fuzz-inputs/temp-9001/merge-workdir/MCF',
        ],
        artifact_prefix=None,
        merge_timeout=1800.0,
        tmp_dir='/fuzz-inputs/temp-9001/merge-workdir')

    self.assertDictEqual({
        'actual_duration': 2,
        'average_exec_per_sec': 21,
        'bad_instrumentation': 0,
        'corpus_crash_count': 0,
        'corpus_size': 0,
        'crash_count': 1,
        'dict_used': 1,
        'edge_coverage': 411,
        'edges_total': 398467,
        'expected_duration': 3580,
        'feature_coverage': 1873,
        'fuzzing_time_percent': 0.055865921787709494,
        'initial_edge_coverage': 410,
        'initial_feature_coverage': 1869,
        'leak_count': 0,
        'log_lines_from_engine': 2,
        'log_lines_ignored': 67,
        'log_lines_unwanted': 0,
        'manual_dict_size': 0,
        'max_len': 9001,
        'merge_edge_coverage': 0,
        'new_edges': 1,
        'new_features': 4,
        'new_units_added': 1,
        'new_units_generated': 0,
        'number_of_executed_units': 1249,
        'oom_count': 0,
        'peak_rss_mb': 1197,
        'recommended_dict_size': 0,
        'slow_unit_count': 0,
        'slow_units_count': 0,
        'slowest_unit_time_sec': 0,
        'startup_crash_count': 0,
        'strategy_corpus_mutations_ml_rnn': 0,
        'strategy_corpus_mutations_radamsa': 0,
        'strategy_corpus_subset': 0,
        'strategy_dataflow_tracing': 0,
        'strategy_fork': 0,
        'strategy_mutator_plugin': 0,
        'strategy_mutator_plugin_radamsa': 0,
        'strategy_peach_grammar_mutation': '',
        'strategy_random_max_len': 0,
        'strategy_recommended_dict': 0,
        'strategy_selection_method': 'default',
        'strategy_value_profile': 0,
        'timeout_count': 0,
        'timeout_limit': 123,
    }, result.stats)


def set_strategy_pool(strategies=None):
  """Helper method to create instances of strategy pools
  for patching use."""
  strategy_pool = strategy_selection.StrategyPool()

  if strategies is not None:
    for strategy_tuple in strategies:
      strategy_pool.add_strategy(strategy_tuple)

  return strategy_pool


def mock_random_choice(seq):
  """Always returns first element from the sequence."""
  # We could try to mock a particular |seq| to be a list with a single element,
  # but it does not work well, as random_choice returns a 'mock.mock.MagicMock'
  # object that behaves differently from the actual type of |seq[0]|.
  return seq[0]


def clear_temp_dir():
  """Clear temp directory."""
  if os.path.exists(TEMP_DIR):
    shutil.rmtree(TEMP_DIR)

  os.mkdir(TEMP_DIR)


def setup_testcase_and_corpus(testcase, corpus):
  """Setup testcase and corpus."""
  clear_temp_dir()
  copied_testcase_path = os.path.join(TEMP_DIR, testcase)
  shutil.copy(os.path.join(DATA_DIR, testcase), copied_testcase_path)

  copied_corpus_path = os.path.join(TEMP_DIR, corpus)
  src_corpus_path = os.path.join(DATA_DIR, corpus)

  if os.path.exists(src_corpus_path):
    shutil.copytree(src_corpus_path, copied_corpus_path)
  else:
    os.mkdir(copied_corpus_path)

  return copied_testcase_path, copied_corpus_path


def get_fuzz_timeout(fuzz_time):
  """Return timeout for fuzzing."""
  return (fuzz_time + libfuzzer.LibFuzzerCommon.LIBFUZZER_CLEAN_EXIT_TIME +
          libfuzzer.LibFuzzerCommon.SIGTERM_WAIT_TIME)


def mock_get_directory_file_count(dir_path):
  """Mocked version, always return 1 for new testcases directory."""
  if dir_path == os.path.join(fuzzer_utils.get_temp_dir(), 'new'):
    return 1

  return _get_directory_file_count_orig(dir_path)


class BaseIntegrationTest(unittest.TestCase):
  """Base integration tests."""

  def setUp(self):
    self.maxDiff = None  # pylint: disable=invalid-name
    test_helpers.patch_environ(self)

    os.environ['BUILD_DIR'] = DATA_DIR
    os.environ['FAIL_RETRIES'] = '1'
    os.environ['FUZZ_INPUTS_DISK'] = TEMP_DIR
    os.environ['FUZZ_TEST_TIMEOUT'] = '4800'
    os.environ['TEST_TIMEOUT'] = '65'
    os.environ['JOB_NAME'] = 'libfuzzer_asan'
    os.environ['INPUT_DIR'] = TEMP_DIR
    os.environ['CACHE_DIR'] = TEMP_DIR

    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.dictionary_manager.DictionaryManager.'
        'update_recommended_dictionary',
        'clusterfuzz._internal.bot.fuzzers.engine_common.get_merge_timeout',
        'clusterfuzz._internal.bot.fuzzers.engine_common.random_choice',
        'clusterfuzz._internal.bot.fuzzers.mutator_plugin._download_mutator_plugin_archive',
        'clusterfuzz._internal.bot.fuzzers.mutator_plugin._get_mutator_plugins_from_bucket',
        'clusterfuzz._internal.bot.fuzzers.strategy_selection.'
        'generate_weighted_strategy_pool',
        'clusterfuzz._internal.bot.fuzzers.libfuzzer.get_dictionary_analysis_timeout',
        'clusterfuzz._internal.bot.fuzzers.libfuzzer.get_fuzz_timeout',
        'os.getpid',
        'clusterfuzz._internal.system.minijail.MinijailChroot._mknod',
    ])

    self.mock.getpid.return_value = 1337

    self.mock._get_mutator_plugins_from_bucket.return_value = []  # pylint: disable=protected-access
    self.mock.generate_weighted_strategy_pool.return_value = set_strategy_pool()
    self.mock.get_dictionary_analysis_timeout.return_value = 5
    self.mock.get_merge_timeout.return_value = 10
    self.mock.random_choice.side_effect = mock_random_choice


@test_utils.integration
class IntegrationTests(BaseIntegrationTest):
  """Base libFuzzer libfuzzer tests."""

  def setUp(self):
    BaseIntegrationTest.setUp(self)
    self.crash_dir = TEMP_DIR

  def compare_arguments(self, target_path, arguments, corpora_or_testcase,
                        actual):
    """Compare expected arguments."""
    self.assertListEqual(actual,
                         [target_path] + arguments + corpora_or_testcase)

  def assert_has_stats(self, stats):
    """Asserts that libFuzzer stats are in output."""
    self.assertIn('number_of_executed_units', stats)
    self.assertIn('average_exec_per_sec', stats)
    self.assertIn('new_units_added', stats)
    self.assertIn('slowest_unit_time_sec', stats)
    self.assertIn('peak_rss_mb', stats)

  def test_single_testcase_crash(self):
    """Tests libfuzzer with a crashing testcase."""
    testcase_path, _ = setup_testcase_and_corpus('crash', 'empty_corpus')
    engine_impl = engine.LibFuzzerEngine()
    target_path = engine_common.find_fuzzer_path(DATA_DIR, 'test_fuzzer')
    result = engine_impl.reproduce(target_path, testcase_path,
                                   ['-timeout=60', '-rss_limit_mb=2560'], 65)
    self.compare_arguments(
        os.path.join(DATA_DIR, 'test_fuzzer'),
        ['-timeout=60', '-rss_limit_mb=2560', '-runs=100'], [testcase_path],
        result.command)
    self.assertIn(
        'ERROR: AddressSanitizer: SEGV on unknown address 0x000000000000',
        result.output)

  @test_utils.slow
  def test_fuzz_no_crash(self):
    """Tests fuzzing (no crash)."""
    self.mock.generate_weighted_strategy_pool.return_value = set_strategy_pool(
        [strategy.VALUE_PROFILE_STRATEGY])

    _, corpus_path = setup_testcase_and_corpus('empty', 'corpus')
    engine_impl = engine.LibFuzzerEngine()

    target_path = engine_common.find_fuzzer_path(DATA_DIR, 'test_fuzzer')
    dict_path = target_path + '.dict'
    options = engine_impl.prepare(corpus_path, target_path, DATA_DIR)

    results = engine_impl.fuzz(target_path, options, TEMP_DIR,
                               get_fuzz_timeout(5.0))
    self.assert_has_stats(results.stats)
    self.compare_arguments(
        os.path.join(DATA_DIR, 'test_fuzzer'), [
            '-max_len=256', '-timeout=25', '-rss_limit_mb=2560',
            '-use_value_profile=1', '-dict=' + dict_path,
            '-artifact_prefix=' + TEMP_DIR + '/', '-max_total_time=5',
            '-print_final_stats=1'
        ], [
            os.path.join(TEMP_DIR, 'temp-1337/new'),
            os.path.join(TEMP_DIR, 'corpus')
        ], results.command)
    self.assertEqual(0, len(results.crashes))

    # New items should've been added to the corpus.
    self.assertNotEqual(0, len(os.listdir(corpus_path)))

    # The incremental stats are not zero as the two step merge was used.
    self.assertNotEqual(0, results.stats['new_edges'])
    self.assertNotEqual(0, results.stats['new_features'])

  @test_utils.slow
  def test_fuzz_no_crash_with_old_libfuzzer(self):
    """Tests fuzzing (no crash) with an old version of libFuzzer."""
    self.mock.generate_weighted_strategy_pool.return_value = set_strategy_pool(
        [strategy.VALUE_PROFILE_STRATEGY])

    _, corpus_path = setup_testcase_and_corpus('empty', 'corpus')
    engine_impl = engine.LibFuzzerEngine()

    target_path = engine_common.find_fuzzer_path(DATA_DIR, 'test_fuzzer_old')
    dict_path = target_path + '.dict'
    options = engine_impl.prepare(corpus_path, target_path, DATA_DIR)

    results = engine_impl.fuzz(target_path, options, TEMP_DIR,
                               get_fuzz_timeout(5))
    self.assert_has_stats(results.stats)
    self.compare_arguments(
        os.path.join(DATA_DIR, 'test_fuzzer_old'), [
            '-max_len=256', '-timeout=25', '-rss_limit_mb=2560',
            '-use_value_profile=1', '-dict=' + dict_path,
            '-artifact_prefix=' + TEMP_DIR + '/', '-max_total_time=5',
            '-print_final_stats=1'
        ], [
            os.path.join(TEMP_DIR, 'temp-1337/new'),
            os.path.join(TEMP_DIR, 'corpus')
        ], results.command)
    self.assertEqual(0, len(results.crashes))

    # New items should've been added to the corpus.
    self.assertNotEqual(0, len(os.listdir(corpus_path)))

    # The incremental stats are zero as the single step merge was used.
    self.assertEqual(0, results.stats['new_edges'])
    self.assertEqual(0, results.stats['new_features'])

  def test_fuzz_crash(self):
    """Tests fuzzing (crash)."""
    _, corpus_path = setup_testcase_and_corpus('empty', 'corpus')
    engine_impl = engine.LibFuzzerEngine()

    target_path = engine_common.find_fuzzer_path(DATA_DIR,
                                                 'always_crash_fuzzer')
    options = engine_impl.prepare(corpus_path, target_path, DATA_DIR)

    results = engine_impl.fuzz(target_path, options, TEMP_DIR,
                               get_fuzz_timeout(5))

    self.assert_has_stats(results.stats)
    self.compare_arguments(
        os.path.join(DATA_DIR, 'always_crash_fuzzer'), [
            '-max_len=100', '-timeout=25', '-rss_limit_mb=2560',
            '-artifact_prefix=' + TEMP_DIR + '/', '-max_total_time=5',
            '-print_final_stats=1'
        ], [
            os.path.join(TEMP_DIR, 'temp-1337/new'),
            os.path.join(TEMP_DIR, 'corpus')
        ], results.command)
    self.assertEqual(1, len(results.crashes))

    self.assertTrue(os.path.exists(results.crashes[0].input_path))
    self.assertEqual(TEMP_DIR, os.path.dirname(results.crashes[0].input_path))
    self.assertEqual(results.logs, results.crashes[0].stacktrace)
    self.assertListEqual([
        '-rss_limit_mb=2560',
        '-timeout=60',
    ], results.crashes[0].reproduce_args)

    self.assertIn('Test unit written to {0}/crash-'.format(self.crash_dir),
                  results.logs)
    self.assertIn(
        'ERROR: AddressSanitizer: SEGV on unknown address '
        '0x000000000000', results.logs)

  def test_fuzz_from_subset(self):
    """Tests fuzzing from corpus subset."""
    self.mock.generate_weighted_strategy_pool.return_value = set_strategy_pool(
        [strategy.CORPUS_SUBSET_STRATEGY])

    _, corpus_path = setup_testcase_and_corpus('empty',
                                               'corpus_with_some_files')

    engine_impl = engine.LibFuzzerEngine()
    target_path = engine_common.find_fuzzer_path(DATA_DIR, 'test_fuzzer')
    dict_path = target_path + '.dict'
    options = engine_impl.prepare(corpus_path, target_path, DATA_DIR)
    results = engine_impl.fuzz(target_path, options, TEMP_DIR,
                               get_fuzz_timeout(5))

    self.compare_arguments(
        os.path.join(DATA_DIR, 'test_fuzzer'), [
            '-max_len=256', '-timeout=25', '-rss_limit_mb=2560',
            '-dict=' + dict_path, '-artifact_prefix=' + TEMP_DIR + '/',
            '-max_total_time=5', '-print_final_stats=1'
        ], [
            os.path.join(TEMP_DIR, 'temp-1337/new'),
            os.path.join(TEMP_DIR, 'temp-1337/subset')
        ], results.command)

    self.assert_has_stats(results.stats)

  def test_minimize(self):
    """Tests minimize."""
    testcase_path, _ = setup_testcase_and_corpus('aaaa', 'empty_corpus')
    minimize_output_path = os.path.join(TEMP_DIR, 'minimized_testcase')

    engine_impl = engine.LibFuzzerEngine()
    target_path = engine_common.find_fuzzer_path(DATA_DIR,
                                                 'crash_with_A_fuzzer')
    result = engine_impl.minimize_testcase(target_path, [], testcase_path,
                                           minimize_output_path, 120)
    self.assertTrue(result)
    self.assertTrue(os.path.exists(minimize_output_path))
    with open(minimize_output_path) as f:
      result = f.read()
      self.assertEqual('A', result)

  def test_cleanse(self):
    """Tests cleanse."""
    testcase_path, _ = setup_testcase_and_corpus('aaaa', 'empty_corpus')
    cleanse_output_path = os.path.join(TEMP_DIR, 'cleansed_testcase')

    engine_impl = engine.LibFuzzerEngine()
    target_path = engine_common.find_fuzzer_path(DATA_DIR,
                                                 'crash_with_A_fuzzer')
    result = engine_impl.cleanse(target_path, [], testcase_path,
                                 cleanse_output_path, 120)
    self.assertTrue(result)
    self.assertTrue(os.path.exists(cleanse_output_path))
    with open(cleanse_output_path) as f:
      result = f.read()
      self.assertFalse(all(c == 'A' for c in result))

  def test_analyze_dict(self):
    """Tests recommended dictionary analysis."""
    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.dictionary_manager.DictionaryManager.'
        'parse_recommended_dictionary_from_log_lines',
    ])

    self.mock.parse_recommended_dictionary_from_log_lines.return_value = set([
        '"USELESS_0"',
        '"APPLE"',
        '"USELESS_1"',
        '"GINGER"',
        '"USELESS_2"',
        '"BEET"',
        '"USELESS_3"',
    ])

    _, corpus_path = setup_testcase_and_corpus('empty',
                                               'corpus_with_some_files')

    engine_impl = engine.LibFuzzerEngine()
    target_path = engine_common.find_fuzzer_path(DATA_DIR,
                                                 'analyze_dict_fuzzer')
    options = engine_impl.prepare(corpus_path, target_path, DATA_DIR)
    engine_impl.fuzz(target_path, options, TEMP_DIR, get_fuzz_timeout(5))
    expected_recommended_dictionary = set([
        '"APPLE"',
        '"GINGER"',
        '"BEET"',
    ])

    self.assertIn(expected_recommended_dictionary,
                  self.mock.update_recommended_dictionary.call_args[0])

  def test_fuzz_with_mutator_plugin(self):
    """Tests fuzzing with a mutator plugin."""

    os.environ['MUTATOR_PLUGINS_DIR'] = os.path.join(TEMP_DIR,
                                                     'mutator-plugins')
    # TODO(metzman): Remove the old binary and switch the test to the new one.
    fuzz_target_name = 'test_fuzzer_old'
    plugin_archive_name = (
        'custom_mutator_plugin-libfuzzer_asan-test_fuzzer_old.zip')

    # Call before setting up the plugin since this call will erase the directory
    # the plugin is written to.
    _, corpus_path = setup_testcase_and_corpus('empty', 'empty_corpus')
    plugin_archive_path = os.path.join(DATA_DIR, plugin_archive_name)

    self.mock.generate_weighted_strategy_pool.return_value = set_strategy_pool(
        [strategy.MUTATOR_PLUGIN_STRATEGY])
    self.mock._get_mutator_plugins_from_bucket.return_value = [  # pylint: disable=protected-access
        plugin_archive_name
    ]
    self.mock._download_mutator_plugin_archive.return_value = (  # pylint: disable=protected-access
        plugin_archive_path)
    custom_mutator_print_string = 'CUSTOM MUTATOR\n'
    try:
      target_path = engine_common.find_fuzzer_path(DATA_DIR, fuzz_target_name)
      engine_impl = engine.LibFuzzerEngine()
      options = engine_impl.prepare(corpus_path, target_path, DATA_DIR)
      results = engine_impl.fuzz(target_path, options, TEMP_DIR,
                                 get_fuzz_timeout(5))
    finally:
      shutil.rmtree(os.environ['MUTATOR_PLUGINS_DIR'])

    # custom_mutator_print_string gets printed before the custom mutator mutates
    # a test case. Assert that the count is greater than 1 to ensure that the
    # function didn't crash on its first execution (after printing).
    self.assertGreater(results.logs.count(custom_mutator_print_string), 1)

  def test_merge_reductions(self):
    """Tests that reduced testcases are merged back into the original corpus
    without deleting the larger version."""
    _, corpus_path = setup_testcase_and_corpus('empty', 'empty_corpus')
    fuzz_target_name = 'analyze_dict_fuzzer'

    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.libFuzzer.engine.LibFuzzerEngine.'
        '_create_merge_corpus_dir',
        'clusterfuzz._internal.system.shell.get_directory_file_count',
    ])

    self.mock.get_directory_file_count.side_effect = (
        mock_get_directory_file_count)

    minimal_unit_contents = 'APPLE'
    minimal_unit_hash = '569bea285d70dda2218f89ef5454ea69fb5111ef'
    nonminimal_unit_contents = 'APPLEO'
    nonminimal_unit_hash = '07aef0e305db0779f3b52ab4dad975a1b737c461'

    def mocked_create_merge_directory(_):
      """A mocked version of create_merge_directory that adds some interesting
      files to the merge corpus and initial corpus."""
      merge_directory_path = libfuzzer.create_corpus_directory('merge-corpus')

      # Write the minimal unit to the new corpus directory.
      new_corpus_directory_path = libfuzzer.create_corpus_directory('new')
      minimal_unit_path = os.path.join(new_corpus_directory_path,
                                       minimal_unit_hash)

      with open(minimal_unit_path, 'w+') as file_handle:
        file_handle.write(minimal_unit_contents)

      # Write the nonminimal unit to the corpus directory.
      nonminimal_unit_path = os.path.join(corpus_path, nonminimal_unit_hash)
      with open(nonminimal_unit_path, 'w+') as file_handle:
        file_handle.write(nonminimal_unit_contents)

      return merge_directory_path

    # pylint: disable=protected-access
    self.mock._create_merge_corpus_dir.side_effect = (
        mocked_create_merge_directory)

    target_path = engine_common.find_fuzzer_path(DATA_DIR, fuzz_target_name)
    engine_impl = engine.LibFuzzerEngine()
    options = engine_impl.prepare(corpus_path, target_path, DATA_DIR)
    options.arguments.append('-runs=10')
    engine_impl.fuzz(target_path, options, TEMP_DIR, get_fuzz_timeout(5))

    # Verify that both the newly found minimal testcase and the nonminimal
    # testcase are in the corpus.
    self.assertIn(minimal_unit_hash, os.listdir(corpus_path))
    self.assertIn(nonminimal_unit_hash, os.listdir(corpus_path))

  def test_exit_failure_logged(self):
    """Test that we log when libFuzzer's exit code indicates it ran into an
    error."""
    test_helpers.patch(self, [
        'clusterfuzz._internal.metrics.logs.log_error',
    ])

    def mocked_log_error(*args, **kwargs):  # pylint: disable=unused-argument
      self.assertIn(engine.ENGINE_ERROR_MESSAGE, args[0])

    self.mock.log_error.side_effect = mocked_log_error
    _, corpus_path = setup_testcase_and_corpus('empty',
                                               'corpus_with_some_files')

    target_path = engine_common.find_fuzzer_path(DATA_DIR, 'exit_fuzzer')
    engine_impl = engine.LibFuzzerEngine()
    options = engine_impl.prepare(corpus_path, target_path, DATA_DIR)
    options.extra_env['EXIT_FUZZER_CODE'] = '1'

    results = engine_impl.fuzz(target_path, options, TEMP_DIR,
                               get_fuzz_timeout(5))
    self.assertEqual(1, self.mock.log_error.call_count)

    self.assertEqual(1, len(results.crashes))
    self.assertEqual(TEMP_DIR, os.path.dirname(results.crashes[0].input_path))
    self.assertEqual(0, os.path.getsize(results.crashes[0].input_path))

  @parameterized.parameterized.expand(['77', '27'])
  def test_exit_target_bug_not_logged(self, exit_code):
    """Test that we don't log when exit code indicates bug found in target."""
    test_helpers.patch(self, [
        'clusterfuzz._internal.metrics.logs.log_error',
    ])

    def mocked_log_error(*args, **kwargs):  # pylint: disable=unused-argument
      self.assertNotIn(engine.ENGINE_ERROR_MESSAGE, args[0])

    self.mock.log_error.side_effect = mocked_log_error
    _, corpus_path = setup_testcase_and_corpus('empty',
                                               'corpus_with_some_files')

    target_path = engine_common.find_fuzzer_path(DATA_DIR, 'exit_fuzzer')
    engine_impl = engine.LibFuzzerEngine()
    options = engine_impl.prepare(corpus_path, target_path, DATA_DIR)
    options.extra_env['EXIT_FUZZER_CODE'] = exit_code

    results = engine_impl.fuzz(target_path, options, TEMP_DIR,
                               get_fuzz_timeout(5))

    self.assertEqual(1, len(results.crashes))
    self.assertEqual(TEMP_DIR, os.path.dirname(results.crashes[0].input_path))
    self.assertEqual(0, os.path.getsize(results.crashes[0].input_path))

  def test_fuzz_invalid_dict(self):
    """Tests fuzzing with an invalid dictionary (ParseDictionaryFile crash)."""
    test_helpers.patch(self, [
        'clusterfuzz._internal.metrics.logs.log_error',
    ])

    def mocked_log_error(*args, **kwargs):  # pylint: disable=unused-argument
      self.assertIn('Dictionary parsing failed (target=test_fuzzer, line=2).',
                    args[0])

    self.mock.log_error.side_effect = mocked_log_error
    _, corpus_path = setup_testcase_and_corpus('empty', 'corpus')
    engine_impl = engine.LibFuzzerEngine()

    target_path = engine_common.find_fuzzer_path(DATA_DIR, 'test_fuzzer')
    options = engine_impl.prepare(corpus_path, target_path, DATA_DIR)

    invalid_dict_path = os.path.join(DATA_DIR, 'invalid.dict')
    options.arguments.append('-dict=' + invalid_dict_path)

    engine_impl.fuzz(target_path, options, TEMP_DIR, get_fuzz_timeout(5))


@unittest.skip('needs root')
@test_utils.integration
class UnshareIntegrationTests(IntegrationTests):
  """Unshare runner integration tests."""

  def setUp(self):
    super().setUp()
    os.environ['USE_UNSHARE'] = 'True'

  def compare_arguments(self, target_path, arguments, corpora_or_testcase,
                        actual):
    """Compare expected arguments."""
    self.assertListEqual(actual, ['/usr/bin/unshare', '-n', target_path] +
                         arguments + corpora_or_testcase)


@test_utils.integration
class MinijailIntegrationTests(IntegrationTests):
  """Minijail integration tests."""

  def setUp(self):
    IntegrationTests.setUp(self)
    os.environ['USE_MINIJAIL'] = 'True'
    self.crash_dir = '/temp'

  def compare_arguments(self, target_path, arguments, corpora_or_testcase,
                        actual):
    """Overridden compare_arguments."""

    def _to_chroot_path(path):
      """Convert to chroot path."""
      return '/' + os.path.basename(path.rstrip('/'))

    for i, argument in enumerate(arguments):
      if not argument.startswith(constants.ARTIFACT_PREFIX_FLAG):
        continue

      arguments[i] = constants.ARTIFACT_PREFIX_FLAG + _to_chroot_path(
          argument[len(constants.ARTIFACT_PREFIX_FLAG):]) + '/'

    expected_arguments = [target_path] + arguments + [
        _to_chroot_path(item) for item in corpora_or_testcase
    ]
    # Ignore minijail arguments
    self.assertListEqual(expected_arguments, actual[-len(expected_arguments):])

  def test_exit_failure_logged(self):
    """Exit failure is not logged in minijail."""

  @parameterized.parameterized.expand(['1', '77', '27'])
  def test_exit_target_bug_not_logged(self, exit_code):
    """Test that we don't log when exit code indicates bug found in target."""
    test_helpers.patch(self, [
        'clusterfuzz._internal.metrics.logs.log_error',
    ])

    def mocked_log_error(*args, **kwargs):  # pylint: disable=unused-argument
      self.assertNotIn(engine.ENGINE_ERROR_MESSAGE, args[0])

    self.mock.log_error.side_effect = mocked_log_error
    _, corpus_path = setup_testcase_and_corpus('empty',
                                               'corpus_with_some_files')

    target_path = engine_common.find_fuzzer_path(DATA_DIR, 'exit_fuzzer')
    engine_impl = engine.LibFuzzerEngine()
    options = engine_impl.prepare(corpus_path, target_path, DATA_DIR)
    options.extra_env['EXIT_FUZZER_CODE'] = exit_code

    results = engine_impl.fuzz(target_path, options, TEMP_DIR,
                               get_fuzz_timeout(5))

    self.assertEqual(1, len(results.crashes))
    self.assertEqual(TEMP_DIR, os.path.dirname(results.crashes[0].input_path))
    self.assertEqual(0, os.path.getsize(results.crashes[0].input_path))


@test_utils.integration
@test_utils.with_cloud_emulators('datastore')
class IntegrationTestsFuchsia(BaseIntegrationTest):
  """libFuzzer tests (Fuchsia)."""

  def setUp(self):
    BaseIntegrationTest.setUp(self)
    self.temp_dir = tempfile.mkdtemp()
    builds_dir = os.path.join(self.temp_dir, 'builds')
    os.mkdir(builds_dir)
    urls_dir = os.path.join(self.temp_dir, 'urls')
    os.mkdir(urls_dir)

    environment.set_value('BUILDS_DIR', builds_dir)
    environment.set_value('BUILD_URLS_DIR', urls_dir)
    environment.set_value('QUEUE_OVERRIDE', 'FUCHSIA')
    environment.set_value('OS_OVERRIDE', 'FUCHSIA')
    environment.set_value(
        'RELEASE_BUILD_BUCKET_PATH',
        'gs://clusterfuchsia-builds-test/libfuzzer/'
        'fuchsia-([0-9]+).zip')
    environment.set_value('UNPACK_ALL_FUZZ_TARGETS_AND_FILES', True)
    test_helpers.patch(self, [
        'clusterfuzz._internal.system.shell.clear_temp_directory',
    ])

  def tearDown(self):
    shutil.rmtree(self.temp_dir, ignore_errors=True)

  @unittest.skipIf(
      not environment.get_value('FUCHSIA_TESTS'),
      'Temporarily disabling the Fuchsia test until build size reduced.')
  def test_fuzzer_can_boot_and_run_with_corpus(self):
    """Tests running a single round of fuzzing on a Fuchsia target, using
    a toy fuzzer that should crash very quickly.

    Additionally, tests that pushing a corpus to the target works & produces
    an expanded corpus."""
    environment.set_value('JOB_NAME', 'libfuzzer_asan_fuchsia')
    environment.set_value('FUZZ_TARGET', 'example-fuzzers/crash_fuzzer')
    build_manager.setup_build()

    _, corpus_path = setup_testcase_and_corpus('aaaa', 'fuchsia_corpus')
    num_files_original = len(os.listdir(corpus_path))
    engine_impl = engine.LibFuzzerEngine()

    options = engine_impl.prepare(corpus_path, 'example-fuzzers/crash_fuzzer',
                                  DATA_DIR)
    results = engine_impl.fuzz('example-fuzzers/crash_fuzzer', options,
                               TEMP_DIR, get_fuzz_timeout(20))

    # If we don't get a crash, something went wrong.
    self.assertIn('Test unit written to', results.logs)
    # Check that the command was invoked with a corpus argument.
    self.assertIn('data/corpus/new', results.command)
    # Check that new units were added to the corpus.
    num_files_new = len(os.listdir(os.path.join(TEMP_DIR, 'temp-1337/new')))
    self.assertGreater(num_files_new, num_files_original)

  @unittest.skipIf(
      not environment.get_value('FUCHSIA_TESTS'),
      'Temporarily disabling the Fuchsia tests until build size reduced.')
  def test_fuzzer_can_boot_and_run_reproducer(self):
    """Tests running a testcase that should cause a fast, predictable crash."""
    environment.set_value('FUZZ_TARGET', 'example-fuzzers/overflow_fuzzer')
    environment.set_value('JOB_NAME', 'libfuzzer_asan_fuchsia')
    build_manager.setup_build()
    testcase_path, _ = setup_testcase_and_corpus('fuchsia_crash',
                                                 'empty_corpus')
    engine_impl = engine.LibFuzzerEngine()
    result = engine_impl.reproduce('example-fuzzers/overflow_fuzzer',
                                   testcase_path,
                                   ['-timeout=25', '-rss_limit_mb=2560'], 30)

    self.assertIn('ERROR: AddressSanitizer: heap-buffer-overflow on address',
                  result.output)
    self.assertIn('Running: data/fuchsia_crash', result.output)

  @unittest.skipIf(
      not environment.get_value('FUCHSIA_TESTS'),
      'Temporarily disabling the Fuchsia tests until build size reduced.')
  def test_qemu_logs_returned_on_error(self):
    """Test running against a qemu that has died"""
    test_helpers.patch(self, ['clusterfuzz._internal.metrics.logs.log_warn'])
    # Pass-through logs just so we can see what's going on (but moving from
    # log_warn to plain log to avoid creating a loop)
    self.mock.log_warn.side_effect = logs.log

    environment.set_value('FUZZ_TARGET', 'example-fuzzers/crash_fuzzer')
    environment.set_value('JOB_NAME', 'libfuzzer_asan_fuchsia')
    build_manager.setup_build()
    testcase_path, _ = setup_testcase_and_corpus('fuchsia_crash',
                                                 'empty_corpus')

    engine_impl = engine.LibFuzzerEngine()

    # Check that it's up properly
    results = engine_impl.reproduce('example-fuzzers/overflow_fuzzer',
                                    testcase_path,
                                    ['-timeout=25', '-rss_limit_mb=2560'], 30)
    self.assertEqual(0, results.return_code)

    # Force termination
    process_handler.terminate_processes_matching_names('qemu-system-x86_64')

    # Try to fuzz against the dead qemu to trigger log dump (and automatic
    # recovery behavior, when undercoat is disabled)
    try:
      engine_impl.reproduce('example-fuzzers/overflow_fuzzer', testcase_path,
                            ['-timeout=25', '-rss_limit_mb=2560'], 30)
    except:
      # With undercoat, this is expected to dump logs but ultimately fail to run
      if not environment.get_value('FUCHSIA_USE_UNDERCOAT'):
        raise

    # Check the logs for syslog presence
    self.assertIn('{{{reset}}}', self.mock.log_warn.call_args[0][0])

  @unittest.skipIf(
      not environment.get_value('FUCHSIA_TESTS'),
      'Temporarily disabling the Fuchsia tests until build size reduced.')
  def test_minimize_testcase(self):
    """Tests running a testcase that should be able to minimize."""
    environment.set_value('FUZZ_TARGET', 'example-fuzzers/crash_fuzzer')
    environment.set_value('JOB_NAME', 'libfuzzer_asan_fuchsia')
    build_manager.setup_build()
    testcase_path, _ = setup_testcase_and_corpus('fuchsia_overlong_crash',
                                                 'empty_corpus')
    minimize_output_path = os.path.join(TEMP_DIR, 'output')

    engine_impl = engine.LibFuzzerEngine()
    result = engine_impl.minimize_testcase('example-fuzzers/crash_fuzzer',
                                           ['-runs=1000000'], testcase_path,
                                           minimize_output_path, 30)
    with open(minimize_output_path) as f:
      result = f.read()
      self.assertEqual('HI!', result)


@test_utils.integration
@test_utils.with_cloud_emulators('datastore')
class IntegrationTestsAndroid(BaseIntegrationTest, android_helpers.AndroidTest):
  """libFuzzer tests (Android)."""

  def setUp(self):
    android_helpers.AndroidTest.setUp(self)
    BaseIntegrationTest.setUp(self)

    if android.settings.get_sanitizer_tool_name() != 'hwasan':
      raise Exception('Device is not set up with HWASan.')

    environment.set_value('BUILD_DIR', ANDROID_DATA_DIR)
    environment.set_value('JOB_NAME', 'libfuzzer_hwasan_android_device')
    environment.reset_current_memory_tool_options()

    self.crash_dir = TEMP_DIR
    self.adb_path = android.adb.get_adb_path()
    self.hwasan_options = 'HWASAN_OPTIONS="%s"' % quote(
        environment.get_value('HWASAN_OPTIONS'))

  def device_path(self, local_path):
    """Return device path for a local path."""
    return os.path.join(
        android.constants.DEVICE_FUZZING_DIR,
        os.path.relpath(local_path, environment.get_root_directory()))

  def assert_has_stats(self, stats):
    """Asserts that libFuzzer stats are in output."""
    self.assertIn('number_of_executed_units', stats)
    self.assertIn('average_exec_per_sec', stats)
    self.assertIn('new_units_added', stats)
    self.assertIn('slowest_unit_time_sec', stats)
    self.assertIn('peak_rss_mb', stats)

  def test_single_testcase_crash(self):
    """Tests libfuzzer with a crashing testcase."""
    testcase_path, _ = setup_testcase_and_corpus('crash', 'empty_corpus')
    engine_impl = engine.LibFuzzerEngine()
    target_path = engine_common.find_fuzzer_path(ANDROID_DATA_DIR,
                                                 'test_fuzzer')
    result = engine_impl.reproduce(target_path, testcase_path,
                                   ['-timeout=60', '-rss_limit_mb=2560'], 65)

    self.assertEqual([
        self.adb_path, 'shell', self.hwasan_options,
        self.device_path(target_path), '-timeout=60', '-rss_limit_mb=2560',
        '-runs=100',
        self.device_path(testcase_path)
    ], result.command)
    self.assertIn(
        'ERROR: HWAddressSanitizer: SEGV on unknown address 0x000000000000',
        result.output)

  @test_utils.slow
  def test_fuzz_no_crash(self):
    """Tests fuzzing (no crash)."""
    self.mock.generate_weighted_strategy_pool.return_value = set_strategy_pool(
        [strategy.VALUE_PROFILE_STRATEGY])

    _, corpus_path = setup_testcase_and_corpus('empty', 'corpus')
    engine_impl = engine.LibFuzzerEngine()

    target_path = engine_common.find_fuzzer_path(ANDROID_DATA_DIR,
                                                 'test_fuzzer')
    dict_path = target_path + '.dict'
    options = engine_impl.prepare(corpus_path, target_path, ANDROID_DATA_DIR)

    results = engine_impl.fuzz(target_path, options, TEMP_DIR,
                               get_fuzz_timeout(5))

    self.assert_has_stats(results.stats)
    self.assertEqual([
        self.adb_path,
        'shell',
        self.hwasan_options,
        self.device_path(target_path),
        '-max_len=256',
        '-timeout=25',
        '-rss_limit_mb=2560',
        '-use_value_profile=1',
        '-dict=' + self.device_path(dict_path),
        '-artifact_prefix=' + self.device_path(TEMP_DIR) + '/',
        '-max_total_time=5',
        '-print_final_stats=1',
        self.device_path(os.path.join(TEMP_DIR, 'temp-1337/new')),
        self.device_path(os.path.join(TEMP_DIR, 'corpus')),
    ], results.command)
    self.assertTrue(android.adb.file_exists(self.device_path(dict_path)))
    self.assertEqual(0, len(results.crashes))

    # New items should've been added to the corpus.
    self.assertNotEqual(0, len(os.listdir(corpus_path)))

    self.assertNotIn('HWAddressSanitizer:', results.logs)
    self.assertIn('Logcat:', results.logs)

  def test_fuzz_crash(self):
    """Tests fuzzing (crash)."""
    _, corpus_path = setup_testcase_and_corpus('empty', 'corpus')
    engine_impl = engine.LibFuzzerEngine()

    target_path = engine_common.find_fuzzer_path(ANDROID_DATA_DIR,
                                                 'always_crash_fuzzer')
    options = engine_impl.prepare(corpus_path, target_path, ANDROID_DATA_DIR)

    results = engine_impl.fuzz(target_path, options, TEMP_DIR,
                               get_fuzz_timeout(5))

    self.assert_has_stats(results.stats)
    self.assertEqual([
        self.adb_path,
        'shell',
        self.hwasan_options,
        self.device_path(target_path),
        '-max_len=100',
        '-timeout=25',
        '-rss_limit_mb=2560',
        '-artifact_prefix=' + self.device_path(TEMP_DIR) + '/',
        '-max_total_time=5',
        '-print_final_stats=1',
        self.device_path(os.path.join(TEMP_DIR, 'temp-1337/new')),
        self.device_path(os.path.join(TEMP_DIR, 'corpus')),
    ], results.command)
    self.assertEqual(1, len(results.crashes))

    self.assertTrue(os.path.exists(results.crashes[0].input_path))
    self.assertEqual(TEMP_DIR, os.path.dirname(results.crashes[0].input_path))
    self.assertEqual(results.logs, results.crashes[0].stacktrace)
    self.assertListEqual([
        '-rss_limit_mb=2560',
        '-timeout=60',
    ], results.crashes[0].reproduce_args)

    self.assertIn(
        'Test unit written to {0}/crash-'.format(
            self.device_path(self.crash_dir)), results.logs)
    self.assertIn(
        'ERROR: HWAddressSanitizer: SEGV on unknown address '
        '0x000000000000', results.logs)
    self.assertNotIn('Logcat:', results.logs)

  def test_fuzz_from_subset(self):
    """Tests fuzzing from corpus subset."""
    self.mock.generate_weighted_strategy_pool.return_value = set_strategy_pool(
        [strategy.CORPUS_SUBSET_STRATEGY])

    _, corpus_path = setup_testcase_and_corpus('empty',
                                               'corpus_with_some_files')

    engine_impl = engine.LibFuzzerEngine()
    target_path = engine_common.find_fuzzer_path(ANDROID_DATA_DIR,
                                                 'test_fuzzer')
    dict_path = target_path + '.dict'
    options = engine_impl.prepare(corpus_path, target_path, ANDROID_DATA_DIR)
    results = engine_impl.fuzz(target_path, options, TEMP_DIR,
                               get_fuzz_timeout(5))

    self.assertEqual([
        self.adb_path,
        'shell',
        self.hwasan_options,
        self.device_path(target_path),
        '-max_len=256',
        '-timeout=25',
        '-rss_limit_mb=2560',
        '-dict=' + self.device_path(dict_path),
        '-artifact_prefix=' + self.device_path(TEMP_DIR) + '/',
        '-max_total_time=5',
        '-print_final_stats=1',
        self.device_path(os.path.join(TEMP_DIR, 'temp-1337/new')),
        self.device_path(os.path.join(TEMP_DIR, 'temp-1337/subset')),
    ], results.command)
    self.assertTrue(android.adb.file_exists(self.device_path(dict_path)))
    self.assert_has_stats(results.stats)

  def test_minimize(self):
    """Tests minimize."""
    testcase_path, _ = setup_testcase_and_corpus('aaaa', 'empty_corpus')
    minimize_output_path = os.path.join(TEMP_DIR, 'minimized_testcase')

    engine_impl = engine.LibFuzzerEngine()
    target_path = engine_common.find_fuzzer_path(ANDROID_DATA_DIR,
                                                 'crash_with_A_fuzzer')
    result = engine_impl.minimize_testcase(target_path, [], testcase_path,
                                           minimize_output_path, 120)
    self.assertTrue(result)
    self.assertTrue(os.path.exists(minimize_output_path))
    with open(minimize_output_path) as f:
      result = f.read()
      self.assertEqual('A', result)

  def test_cleanse(self):
    """Tests cleanse."""
    testcase_path, _ = setup_testcase_and_corpus('aaaa', 'empty_corpus')
    cleanse_output_path = os.path.join(TEMP_DIR, 'cleansed_testcase')

    engine_impl = engine.LibFuzzerEngine()
    target_path = engine_common.find_fuzzer_path(ANDROID_DATA_DIR,
                                                 'crash_with_A_fuzzer')
    result = engine_impl.cleanse(target_path, [], testcase_path,
                                 cleanse_output_path, 120)
    self.assertTrue(result)
    self.assertTrue(os.path.exists(cleanse_output_path))
    with open(cleanse_output_path) as f:
      result = f.read()
      self.assertFalse(all(c == 'A' for c in result))

  def test_analyze_dict(self):
    """Tests recommended dictionary analysis."""
    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.dictionary_manager.DictionaryManager.'
        'parse_recommended_dictionary_from_log_lines',
    ])

    self.mock.parse_recommended_dictionary_from_log_lines.return_value = set([
        '"USELESS_0"',
        '"APPLE"',
        '"USELESS_1"',
        '"GINGER"',
        '"USELESS_2"',
        '"BEET"',
        '"USELESS_3"',
    ])

    _, corpus_path = setup_testcase_and_corpus('empty',
                                               'corpus_with_some_files')

    engine_impl = engine.LibFuzzerEngine()
    target_path = engine_common.find_fuzzer_path(ANDROID_DATA_DIR,
                                                 'analyze_dict_fuzzer')
    options = engine_impl.prepare(corpus_path, target_path, DATA_DIR)

    engine_impl.fuzz(target_path, options, TEMP_DIR, get_fuzz_timeout(5.0))
    expected_recommended_dictionary = set([
        '"APPLE"',
        '"GINGER"',
        '"BEET"',
    ])

    self.assertIn(expected_recommended_dictionary,
                  self.mock.update_recommended_dictionary.call_args[0])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Gestures related functions."""

import ast
import random
import time

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.metrics import logs

try:
  from pywinauto import application
except ImportError:
  # This can be imported from appengine, so make sure we don't exception out.
  pass

RELOAD_GESTURE = 'key,{F5}'


def find_windows_for_process(process_id):
  """Return visible windows belonging to a process."""
  pids = utils.get_process_ids(process_id)
  if not pids:
    return []

  visible_windows = []
  for pid in pids:
    app = application.Application()
    try:
      app.connect(process=pid)
    except:
      logs.log_warn('Unable to connect to process.')
      continue

    try:
      windows = app.windows()
    except:
      logs.log_warn('Unable to get application windows.')
      continue

    for window in windows:
      try:
        window.type_keys('')
      except:
        continue

      visible_windows.append(window)

  return visible_windows


def get_random_gestures(gesture_count):
  """Return list of random gesture command strings."""
  gestures_types = [
      'key,Letters', 'key,Letters', 'key,Letters', 'key,Letters', 'mouse,MA',
      'mousedrag,MB', 'mousemove,MC'
  ]

  gestures = []
  for _ in range(gesture_count):
    random_gesture = utils.random_element_from_list(gestures_types)
    if 'Letters' in random_gesture:
      num_letters = random.randint(1, 10)
      letters = []
      for _ in range(num_letters):
        if not random.randint(0, 7):
          letters.append(
              utils.random_element_from_list([
                  '{BACK}', '{BACKSPACE}', '{BKSP}', '{CAP}', '{DEL}',
                  '{DELETE}', '{DOWN}', '{DOWN}', '{END}', '{ENTER}', '{ENTER}',
                  '{ENTER}', 'A{ESC}', '{F1}', '{F2}', '{F3}', 'A{F4}', '{F5}',
                  '{F6}', '{F7}', '{F8}', '{F9}', '{F10}', '{F11}', '{F12}',
                  '{HOME}', '{INSERT}', '{LEFT}', '{PGDN}', '{PGUP}', '{RIGHT}',
                  '{SPACE}', '{TAB}', '{TAB}', '{TAB}', '{TAB}', '{UP}', '{UP}',
                  '+', '^'
              ]))
        else:
          letters.append(
              utils.random_element_from_list(
                  ['{TAB}', '^=', '^-',
                   '{%s}' % chr(random.randint(32, 126))]))
      random_gesture = random_gesture.replace('Letters', ''.join(letters))
      if ('^c' in random_gesture.lower() or '^d' in random_gesture.lower() or
          '^z' in random_gesture.lower()):
        continue

    if ',MA' in random_gesture:
      button = utils.random_element_from_list(['left', 'right', 'middle'])
      coords = '(%d,%d)' % (random.randint(0, 1000), random.randint(0, 1000))
      double = utils.random_element_from_list(['True', 'False'])
      random_gesture = random_gesture.replace(
          'MA', '%s;%s;%s' % (button, coords, double))

    if ',MB' in random_gesture:
      button = utils.random_element_from_list(['left', 'right', 'middle'])
      coords1 = '(%d,%d)' % (random.randint(0, 1000), random.randint(0, 1000))
      coords2 = '(%d,%d)' % (random.randint(0, 1000), random.randint(0, 1000))
      random_gesture = random_gesture.replace(
          'MB', '%s;%s;%s' % (button, coords1, coords2))

    if ',MC' in random_gesture:
      button = utils.random_element_from_list(['left', 'right', 'middle'])
      coords = '(%d,%d)' % (random.randint(0, 1000), random.randint(0, 1000))
      random_gesture = random_gesture.replace('MC', '%s;%s' % (button, coords))

    gestures.append(random_gesture)

  return gestures


def run_gestures(gestures, process_id, process_status, start_time, timeout,
                 windows):
  """Run the provided interaction gestures."""
  if not windows:
    windows += find_windows_for_process(process_id)

  for window in windows:
    for gesture in gestures:
      # If process had exited or our timeout interval has exceeded,
      # just bail out.
      if process_status.finished or time.time() - start_time >= timeout:
        return

      try:
        tokens = gesture.split(',')
        command = tokens.pop(0)
        value = ','.join(tokens)

        if command == 'key':
          window.type_keys(value)

        elif command == 'mouse':
          button, coords, double = value.split(';')
          window.click_input(
              button=button,
              coords=ast.literal_eval(coords),
              double=ast.literal_eval(double))

        elif command == 'mousedrag':
          button, coords1, coords2 = value.split(';')
          window.drag_mouse(
              button=button,
              press_coords=ast.literal_eval(coords1),
              release_coords=ast.literal_eval(coords2))

        elif command == 'mousemove':
          button, coords = value.split(';')
          window.move_mouse(pressed=button, coords=ast.literal_eval(coords))

      except Exception:
        # Several types of errors can happen. Just ignore them until a better
        # solution is available. E.g. controls not visible, gestures cannot be
        # run, invalid window handle, window failed to respond to gesture, etc.
        pass



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Retry tests."""
import unittest

# pylint: disable=protected-access
import mock

from clusterfuzz._internal.base import retry
from clusterfuzz._internal.metrics import monitor
from clusterfuzz._internal.metrics import monitoring_metrics
from clusterfuzz._internal.tests.test_libs import helpers


class WrapTest(unittest.TestCase):
  """Test retry decorator."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.base.retry.sleep',
    ])

    self.func_body = mock.MagicMock()
    monitor.metrics_store().reset_for_testing()

  @retry.wrap(retries=4, delay=10, backoff=2, function='_func')
  def _func(self, a):
    return self.func_body(a)

  @retry.wrap(retries=4, delay=10, backoff=2, function='_func')
  def _yield_func(self, a):
    for _ in range(3):
      yield self.func_body(a)

  @retry.wrap(
      retries=4, delay=10, backoff=2, function='_func', retry_on_false=True)
  def _func2(self, a):
    return self.func_body(a)

  class _FakeException(Exception):
    pass

  @retry.wrap(
      retries=4,
      delay=10,
      backoff=2,
      function='_func',
      exception_type=_FakeException)
  def _func_exception_type(self, a):
    return self.func_body(a)

  @retry.wrap(
      retries=4,
      delay=10,
      backoff=2,
      function='_func',
      exception_type=_FakeException)
  def _yield_func_exception_type(self, a):
    for _ in range(3):
      yield self.func_body(a)

  def test_retry_and_succeed(self):
    """Test when retry once and succeed for regular function.."""
    self.func_body.side_effect = [self._FakeException(), 456]

    self.assertEqual(456, self._func(123))

    self.assertEqual(2, self.func_body.call_count)
    self.func_body.assert_has_calls([mock.call(123), mock.call(123)])

    self.assertEqual(1, self.mock.sleep.call_count)
    self.mock.sleep.assert_has_calls([mock.call(10)])

    self.assertEqual(
        1,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': True,
        }))
    self.assertEqual(
        0,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': False,
        }))

  def test_retry_and_succeed_yield(self):
    """Test when retry once and succeed for generator function.."""
    self.func_body.side_effect = [self._FakeException(), 1, 2, 3]

    results = list(self._yield_func(123))
    self.assertEqual([1, 2, 3], results)

    self.assertEqual(4, self.func_body.call_count)
    self.func_body.assert_has_calls(
        [mock.call(123),
         mock.call(123),
         mock.call(123),
         mock.call(123)])

    self.assertEqual(1, self.mock.sleep.call_count)
    self.mock.sleep.assert_has_calls([mock.call(10)])

    self.assertEqual(
        1,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': True,
        }))
    self.assertEqual(
        0,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': False,
        }))

  def test_retry_and_succeed_yield_with_exceptions_in_middle(self):
    """Test when retry once and succeed for generator function with exceptions
    happening after the first element."""
    self.func_body.side_effect = [
        1, self._FakeException(), 1, 2,
        self._FakeException(), 1, 2, 3
    ]

    results = list(self._yield_func(123))
    self.assertEqual([1, 2, 3], results)

    self.assertEqual(8, self.func_body.call_count)
    self.func_body.assert_has_calls([mock.call(123)] * 8)

    self.assertEqual(2, self.mock.sleep.call_count)
    self.mock.sleep.assert_has_calls([mock.call(10), mock.call(20)])

    self.assertEqual(
        1,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': True,
        }))
    self.assertEqual(
        0,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': False,
        }))

  def test_exceed_try_limit(self):
    """Test when exceeding limit for regular function."""
    self.func_body.side_effect = self._FakeException()

    with self.assertRaises(self._FakeException):
      self._func(123)

    self.assertEqual(5, self.func_body.call_count)
    self.func_body.assert_has_calls([mock.call(123)] * 5)

    self.assertEqual(4, self.mock.sleep.call_count)
    self.mock.sleep.assert_has_calls(
        [mock.call(10),
         mock.call(20),
         mock.call(40),
         mock.call(80)])

    self.assertEqual(
        0,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': True,
        }))
    self.assertEqual(
        1,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': False,
        }))

  def test_exceed_try_limit_yield(self):
    """Test when exceeding limit for generator function."""
    self.func_body.side_effect = self._FakeException()

    with self.assertRaises(self._FakeException):
      for _ in self._yield_func(123):
        pass

    self.assertEqual(5, self.func_body.call_count)
    self.func_body.assert_has_calls([mock.call(123)] * 5)

    self.assertEqual(4, self.mock.sleep.call_count)
    self.mock.sleep.assert_has_calls(
        [mock.call(10),
         mock.call(20),
         mock.call(40),
         mock.call(80)])

    self.assertEqual(
        0,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': True,
        }))
    self.assertEqual(
        1,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': False,
        }))

  def test_retry_exception_type_mismatch(self):
    """Test retry with exception mismatching type for regular function."""
    self.func_body.side_effect = [Exception]

    with self.assertRaises(Exception):
      self._func_exception_type(123)

    self.assertEqual(1, self.func_body.call_count)
    self.assertEqual(
        0,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': True,
        }))
    self.assertEqual(
        1,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': False,
        }))

  def test_retry_exception_type_mismatch_yield(self):
    """Test retry with exception mismatching type for generator function."""
    self.func_body.side_effect = [Exception]

    with self.assertRaises(Exception):
      for _ in self._yield_func_exception_type(123):
        pass

    self.assertEqual(1, self.func_body.call_count)
    self.assertEqual(
        0,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': True,
        }))
    self.assertEqual(
        1,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': False,
        }))

  def test_retry_exception_type_match(self):
    """Test retry with exception matching type for regular function."""
    self.func_body.side_effect = [self._FakeException(), 456]

    self.assertEqual(456, self._func_exception_type(123))
    self.assertEqual(2, self.func_body.call_count)
    self.func_body.assert_has_calls([mock.call(123), mock.call(123)])
    self.assertEqual(
        1,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': True,
        }))
    self.assertEqual(
        0,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': False,
        }))

  def test_retry_exception_type_match_yield(self):
    """Test retry with exception matching type for generator function."""
    self.func_body.side_effect = [self._FakeException(), 1, 2, 3]

    results = list(self._yield_func_exception_type(123))
    self.assertEqual([1, 2, 3], results)
    self.assertEqual(4, self.func_body.call_count)
    self.func_body.assert_has_calls(
        [mock.call(123),
         mock.call(123),
         mock.call(123),
         mock.call(123)])
    self.assertEqual(
        1,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': True,
        }))
    self.assertEqual(
        0,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': False,
        }))

  def test_retry_succeed_on_false(self):
    """Test retry on returning false and succeeding later."""
    self.func_body.side_effect = [False, True]
    self.assertTrue(self._func2(123))
    self.assertEqual(2, self.func_body.call_count)
    self.func_body.assert_has_calls([mock.call(123), mock.call(123)])
    self.assertEqual(1, self.mock.sleep.call_count)
    self.mock.sleep.assert_has_calls([mock.call(10)])

    self.assertEqual(
        1,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': True,
        }))
    self.assertEqual(
        0,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': False,
        }))

  def test_retry_fail_on_false(self):
    """Test retry on returning false."""
    self.func_body.return_value = False
    self.assertFalse(self._func2(123))

    self.assertEqual(5, self.func_body.call_count)
    self.func_body.assert_has_calls([mock.call(123)] * 5)

    self.assertEqual(4, self.mock.sleep.call_count)
    self.mock.sleep.assert_has_calls(
        [mock.call(10),
         mock.call(20),
         mock.call(40),
         mock.call(80)])

    self.assertEqual(
        0,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': True,
        }))
    self.assertEqual(
        1,
        monitoring_metrics.TRY_COUNT.get({
            'function': '_func',
            'is_succeeded': False,
        }))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuzzer statistics handler."""

import datetime
import html
import re
import urllib.parse

from flask import request
from googleapiclient.errors import HttpError
import six
import yaml

from clusterfuzz._internal.base import external_users
from clusterfuzz._internal.base import memoize
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import big_query
from clusterfuzz._internal.metrics import fuzzer_stats
from clusterfuzz._internal.metrics import logs
from handlers import base_handler
from libs import access
from libs import handler
from libs import helpers

# Old fuzzer stats don't change, we could cache forever and it would never go
# stale. Since stats can get pretty large and probably aren't used much the day
# after first accessed, use a TTL reflecting this.
MEMCACHE_OLD_TTL_IN_SECONDS = 24 * 60 * 60

# New fuzzer stats change, and aren't as likely to be reaccessed, don't cache
# for very long.
MEMCACHE_TODAY_TTL_IN_SECONDS = 30 * 60


class QueryField(object):
  """Wrapped fuzzer_stats.QueryField with extra metadata."""

  def __init__(self, field, results_index, field_type, bigquery_type):
    self.field = field
    self.results_index = results_index
    self.field_type = field_type
    self.bigquery_type = bigquery_type.lower()


class BuiltinField(object):
  """Wrapped fuzzer_stats.BuiltinField with extra metadata."""

  def __init__(self, spec, field):
    self.spec = spec
    self.field = field


def _bigquery_type_to_charts_type(typename):
  """Convert bigquery type to charts type."""
  typename = typename.lower()
  if typename in ('integer', 'float'):
    return 'number'

  if typename == 'timestamp':
    return 'date'

  return 'string'


def _python_type_to_charts_type(type_value):
  """Convert bigquery type to charts type."""
  if type_value in (int, float):
    return 'number'

  if type_value == datetime.date:
    return 'date'

  return 'string'


def _parse_date(date_str):
  """Parse YYYY-MM-DD."""
  if not date_str:
    return None

  pattern = re.compile(r'^(\d{4})-(\d{2})-(\d{2})$')
  match = pattern.match(date_str)
  if not match:
    return None

  year, month, day = (int(val) for val in match.groups())
  return datetime.date(year, month, day)


def _parse_stats_column_fields(results, stats_columns, group_by, fuzzer, jobs):
  """Parse stats columns."""
  result = []
  columns = fuzzer_stats.parse_stats_column_fields(stats_columns)

  # Insert first column (group by)
  group_by_field_name = fuzzer_stats.group_by_to_field_name(group_by)
  columns.insert(0, fuzzer_stats.QueryField('j', group_by_field_name, None))

  contexts = {}

  for column in columns:
    if isinstance(column, fuzzer_stats.QueryField):
      key = '%s_%s' % (column.table_alias, column.select_alias)

      for i, field_info in enumerate(results['schema']['fields']):
        # the 'name' field could either be "prefix_fieldname" or simply
        # "fieldname"
        if (field_info['name'] == column.select_alias or
            field_info['name'] == key):
          result.append(
              QueryField(column, i,
                         _bigquery_type_to_charts_type(field_info['type']),
                         field_info['type']))
          break
    elif isinstance(column, fuzzer_stats.BuiltinFieldSpecifier):
      # Builtin field.
      # Create new context if it does not exist.
      field_class = column.field_class()
      if not field_class:
        continue

      context_class = field_class.CONTEXT_CLASS
      context = contexts.setdefault(context_class, context_class(fuzzer, jobs))
      result.append(BuiltinField(column, column.create(context)))

  return result


def _parse_group_by(group_by):
  """Parse group_by value."""
  if group_by == 'by-day':
    return fuzzer_stats.QueryGroupBy.GROUP_BY_DAY
  if group_by == 'by-time':
    return fuzzer_stats.QueryGroupBy.GROUP_BY_TIME
  if group_by == 'by-revision':
    return fuzzer_stats.QueryGroupBy.GROUP_BY_REVISION
  if group_by == 'by-job':
    return fuzzer_stats.QueryGroupBy.GROUP_BY_JOB
  if group_by == 'by-fuzzer':
    return fuzzer_stats.QueryGroupBy.GROUP_BY_FUZZER

  return None


def _get_fuzzer_or_engine(name):
  """Return fuzzer entity, or engine this target is part of."""
  fuzz_target = data_handler.get_fuzz_target(name)
  if fuzz_target:
    name = fuzz_target.engine

  return data_types.Fuzzer.query(data_types.Fuzzer.name == name).get()


def _do_bigquery_query(query):
  """Return results from BigQuery."""
  logs.log(query)
  client = big_query.Client()

  try:
    results = client.raw_query(query, max_results=10000)
  except HttpError as e:
    raise helpers.EarlyExitException(str(e), 500)

  if 'rows' not in results:
    raise helpers.EarlyExitException('No stats.', 404)

  return results


def _parse_stats_column_descriptions(stats_column_descriptions):
  """Parse stats column descriptions."""
  if not stats_column_descriptions:
    return {}

  try:
    result = yaml.safe_load(stats_column_descriptions)
    for key, value in six.iteritems(result):
      result[key] = html.escape(value)

    return result
  except yaml.parser.ParserError:
    logs.log_error('Failed to parse stats column descriptions.')
    return {}


def _build_columns(result, columns):
  """Build columns."""
  for column in columns:
    if isinstance(column, QueryField):
      result['cols'].append({
          'label': column.field.select_alias,
          'type': column.field_type,
      })
    elif isinstance(column, BuiltinField):
      result['cols'].append({
          'label': column.spec.alias or column.spec.name,
          'type': _python_type_to_charts_type(column.field.VALUE_TYPE),
      })


def _try_cast(cell, value_str, cast_function, default_value):
  """Try casting the value_str into cast_function."""
  try:
    cell['v'] = cast_function(value_str)
  except (ValueError, TypeError):
    cell['v'] = default_value
    cell['f'] = '--'


# FIXME: Break logic in this function into simpler helper functions.
def _build_rows(result, columns, rows, group_by):
  """Build rows."""
  for row in rows:
    row_data = []
    first_column_value = None
    for column in columns:
      cell = {}
      if isinstance(column, QueryField):
        value = row['f'][column.results_index]['v']

        if column.field.select_alias == 'time':
          timestamp = float(value)
          time = datetime.datetime.utcfromtimestamp(timestamp)
          first_column_value = first_column_value or time
          cell['v'] = 'Date(%d, %d, %d, %d, %d, %d)' % (
              time.year, time.month - 1, time.day, time.hour, time.minute,
              time.second)
        elif column.field.select_alias == 'date':
          timestamp = float(value)
          date = datetime.datetime.utcfromtimestamp(timestamp).date()
          first_column_value = first_column_value or date
          cell['v'] = 'Date(%d, %d, %d)' % (date.year, date.month - 1, date.day)
        elif column.bigquery_type == 'integer':
          _try_cast(cell, value, int, 0)
        elif column.bigquery_type == 'float':
          # Round all float values to single digits.
          _try_cast(cell, value, lambda s: round(float(s), 1), 0.0)
        else:
          cell['v'] = value

        first_column_value = first_column_value or cell['v']
      elif isinstance(column, BuiltinField):
        data = column.field.get(group_by, first_column_value)
        if data:
          formatted_value = data.value
          if data.link:
            link = (
                _get_cloud_storage_link(data.link)
                if data.link.startswith('gs://') else data.link)
            formatted_value = '<a href="%s">%s</a>' % (link, data.value)

          if data.sort_key is not None:
            cell['v'] = data.sort_key
          else:
            cell['v'] = data.value

          if data.sort_key is not None or data.link:
            cell['f'] = formatted_value
        else:
          cell['v'] = ''
          cell['f'] = '--'

      row_data.append(cell)

    result['rows'].append({'c': row_data})


def _get_cloud_storage_link(bucket_path):
  """Return a clickable link to a cloud storage file given the bucket path."""
  return '/gcs-redirect?' + urllib.parse.urlencode({'path': bucket_path})


def _get_filter_from_job(job):
  """Creates a job filter from |job|."""
  return [str(job)] if job else None


def build_results(fuzzer, jobs, group_by, date_start, date_end):
  """Wrapper around the caching wrappers for _build_results. Decides which of
  those wrappers to call based on how long query should be cached for."""
  datetime_end = _parse_date(date_end)
  if not datetime_end:
    raise helpers.EarlyExitException('Missing end date.', 400)

  if datetime_end < utils.utcnow().date():
    logs.log('Building results for older stats %s %s %s %s %s.' %
             (fuzzer, jobs, group_by, date_start, date_end))

    return _build_old_results(fuzzer, jobs, group_by, date_start, date_end)

  logs.log('Building results for stats including today %s %s %s %s %s.' %
           (fuzzer, jobs, group_by, date_start, date_end))

  return _build_todays_results(fuzzer, jobs, group_by, date_start, date_end)


@memoize.wrap(memoize.Memcache(MEMCACHE_TODAY_TTL_IN_SECONDS))
def _build_todays_results(fuzzer, jobs, group_by, date_start, date_end):
  """Wrapper around _build_results that is intended for use by queries where
  date_end is today. Caches results for 15 minutes."""
  return _build_results(fuzzer, jobs, group_by, date_start, date_end)


@memoize.wrap(memoize.Memcache(MEMCACHE_OLD_TTL_IN_SECONDS))
def _build_old_results(fuzzer, jobs, group_by, date_start, date_end):
  """Wrapper around _build_results that is intended for use by queries where
  date_end is before today. Caches results for 24 hours."""
  return _build_results(fuzzer, jobs, group_by, date_start, date_end)


def _build_results(fuzzer, jobs, group_by, date_start, date_end):
  """Build results."""
  date_start = _parse_date(date_start)
  date_end = _parse_date(date_end)

  if not fuzzer or not group_by or not date_start or not date_end:
    raise helpers.EarlyExitException('Missing params.', 400)

  fuzzer_entity = _get_fuzzer_or_engine(fuzzer)
  if not fuzzer_entity:
    raise helpers.EarlyExitException('Fuzzer not found.', 404)

  if fuzzer_entity.stats_columns:
    stats_columns = fuzzer_entity.stats_columns
  else:
    stats_columns = fuzzer_stats.JobQuery.DEFAULT_FIELDS

  group_by = _parse_group_by(group_by)
  if group_by is None:
    raise helpers.EarlyExitException('Invalid grouping.', 400)

  table_query = fuzzer_stats.TableQuery(fuzzer, jobs, stats_columns, group_by,
                                        date_start, date_end)
  results = _do_bigquery_query(table_query.build())

  is_timeseries = group_by == fuzzer_stats.QueryGroupBy.GROUP_BY_TIME
  result = {
      'cols': [],
      'rows': [],
      'column_descriptions':
          _parse_stats_column_descriptions(
              fuzzer_entity.stats_column_descriptions),
      'is_timeseries':
          is_timeseries
  }

  columns = _parse_stats_column_fields(results, stats_columns, group_by, fuzzer,
                                       jobs)

  # If we are grouping by time and plotting graphs, skip builtin columns.
  if is_timeseries:
    columns = [c for c in columns if not isinstance(c, BuiltinField)]

  _build_columns(result, columns)
  _build_rows(result, columns, results['rows'], group_by)
  return result


def _get_date(date_value, days_ago):
  """Returns |date_value| if it is not empty otherwise returns the date
    |days_ago| number of days ago."""
  if date_value:
    return date_value

  date_datetime = utils.utcnow() - datetime.timedelta(days=days_ago)
  return date_datetime.strftime('%Y-%m-%d')


class Handler(base_handler.Handler):
  """Fuzzer stats main page handler."""

  # pylint: disable=unused-argument
  @handler.unsupported_on_local_server
  @handler.get(handler.HTML)
  def get(self, extra=None):
    """Handle a GET request."""
    if not access.has_access():
      # User is an external user of ClusterFuzz (eg: non-Chrome dev who
      # submitted a fuzzer or someone with a project in OSS-Fuzz).
      user_email = helpers.get_user_email()
      fuzzers_list = external_users.allowed_fuzzers_for_user(
          user_email, include_from_jobs=True, include_parents=True)
      if not fuzzers_list:
        # User doesn't actually have access to any fuzzers.
        raise helpers.AccessDeniedException(
            "You don't have access to any fuzzers.")

    return self.render('fuzzer-stats.html', {})


class LoadFiltersHandler(base_handler.Handler):
  """Load filters handler."""

  @handler.unsupported_on_local_server
  @handler.get(handler.HTML)
  def get(self):
    """Handle a GET request."""
    project = request.get('project')

    if access.has_access():
      # User is an internal user of ClusterFuzz (eg: ClusterFuzz developer).

      # Show all projects in the list, since this allows user to pick another
      # project as needed.
      projects_list = data_handler.get_all_project_names()

      # Filter fuzzers and job list if a project is provided.
      fuzzers_list = (
          data_handler.get_all_fuzzer_names_including_children(
              include_parents=True, project=project))
      jobs_list = data_handler.get_all_job_type_names(project=project)
    else:
      # User is an external user of ClusterFuzz (eg: non-Chrome dev who
      # submitted a fuzzer or someone with a project in OSS-Fuzz).
      user_email = helpers.get_user_email()

      # TODO(aarya): Filter fuzzer and job if |project| is provided.
      fuzzers_list = sorted(
          external_users.allowed_fuzzers_for_user(
              user_email, include_from_jobs=True, include_parents=True))
      if not fuzzers_list:
        # User doesn't actually have access to any fuzzers.
        raise helpers.AccessDeniedException(
            "You don't have access to any fuzzers.")

      jobs_list = sorted(external_users.allowed_jobs_for_user(user_email))
      projects_list = sorted(
          {data_handler.get_project_name(job) for job in jobs_list})

    result = {
        'projects': projects_list,
        'fuzzers': fuzzers_list,
        'jobs': jobs_list,
    }
    return self.render_json(result)


class LoadHandler(base_handler.Handler):
  """Load handler."""

  def _check_user_access_and_get_job_filter(self, fuzzer, job):
    """Check whether the current user has access to stats for the fuzzer or job.
    Returns a job filter that should be applied to the query."""
    access_by_fuzzer_or_job = access.has_access(
        fuzzer_name=fuzzer, job_type=job)
    if access_by_fuzzer_or_job:
      # User has full access to the fuzzer, or the specified job.
      # None means no filters => all jobs.
      return _get_filter_from_job(job)

    if not job:
      # Job not specified and user doesn't have full access to the fuzzer. Check
      # if the user has any allowed jobs and use that as a filter.
      allowed_jobs = external_users.allowed_jobs_for_user(
          helpers.get_user_email())
      if allowed_jobs:
        return allowed_jobs

    raise helpers.AccessDeniedException()

  @handler.post(handler.JSON, handler.JSON)
  def post(self):
    """Handle a POST request."""
    fuzzer = request.get('fuzzer')
    job = request.get('job')
    group_by = request.get('group_by')

    # If date_start is "": because the front end defaults to using a
    # start_date 7 days ago, do the same.
    date_start = _get_date(request.get('date_start'), 7)
    # If date_end is "": don't get today's stats as they may not be
    # available, use yesterdays (as the front end does by default).
    date_end = _get_date(request.get('date_end'), 1)

    job_filter = self._check_user_access_and_get_job_filter(fuzzer, job)
    return self.render_json(
        build_results(fuzzer, job_filter, group_by, date_start, date_end))


class PreloadHandler(base_handler.Handler):
  """Handler for the infrequent task of loading results for expensive stats
  queries that are commonly accessed into the cache."""

  def _get_fuzzer_job_filters(self):
    """Return list of fuzzer-job filter tuples."""
    fuzzer_job_filters = []
    for fuzzer_name in data_types.BUILTIN_FUZZERS:
      fuzzer = data_types.Fuzzer.query(
          data_types.Fuzzer.name == fuzzer_name).get()

      for job in fuzzer.jobs:
        fuzzer_job_filters.append((fuzzer_name, _get_filter_from_job(job)))

      # None job is explicitly added for fuzzer query across all jobs.
      fuzzer_job_filters.append((fuzzer_name, _get_filter_from_job(None)))

    return fuzzer_job_filters

  @handler.cron()
  def get(self):
    """Handle a GET request."""
    date_start = _get_date(None, 7)
    date_end = _get_date(None, 1)

    for fuzzer, job_filter in self._get_fuzzer_job_filters():
      group_by = 'by-fuzzer'
      try:
        build_results(fuzzer, job_filter, group_by, date_start, date_end)
      except Exception as e:
        if 'No stats.' not in repr(e):
          logs.log_error('Failed to preload %s %s %s %s %s.' %
                         (fuzzer, job_filter, group_by, date_start, date_end))

      if not job_filter:
        # Group by job only makes sense for queries that do not specify job.
        group_by = 'by-job'
        try:
          build_results(fuzzer, job_filter, group_by, date_start, date_end)
        except Exception as e:
          if 'No stats.' not in repr(e):
            logs.log_error('Failed to preload %s %s %s %s %s.' %
                           (fuzzer, job_filter, group_by, date_start, date_end))


class RefreshCacheHandler(base_handler.Handler):
  """Refresh cache."""

  @handler.cron()
  def get(self):
    """Handle a GET request."""
    fuzzer_logs_context = fuzzer_stats.FuzzerRunLogsContext()
    fuzz_targets = data_handler.get_fuzz_targets()

    # Cache child fuzzer -> logs bucket mappings.
    for fuzz_target in fuzz_targets:
      # pylint: disable=protected-access,unexpected-keyword-arg
      fuzzer_logs_context._get_logs_bucket_from_fuzzer(
          fuzz_target.fully_qualified_name(), __memoize_force__=True)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for performing remote tasks on linux."""
import os
import re
import time

from fabric import api
from fabric import exceptions
from paramiko import ssh_exception

from local.butler import common as butler_common
from local.butler import package


class Handler(object):
  """Handler for performing linux task."""

  def __init__(self, instance_name, platform, project, zone):
    self.instance_name = instance_name
    self.platform = platform
    self.project = project
    self.zone = zone

    self.staging_source_filename = 'clusterfuzz-source-stage.zip'

  def _abspath(self, path):
    """Get absolute path on host given a path inside the clusterfuzz folder."""
    return self.clusterfuzz_parent_path + '/clusterfuzz/' + path

  def _get_run_bot_pids(self):
    """Get the PIDs of run_bot.py."""
    with api.warn_only():
      output = self._run('ps aux | grep run_bot | grep -v grep')

    pids = []
    for line in output.splitlines():
      line = line.strip()
      if line and self._should_kill(line):
        pids.append(re.split(r'\s+', line)[1])

    return pids

  def _log_path(self, log_name):
    return '{log_dir}/{log_name}.log'.format(
        log_dir=self._abspath('bot/logs'), log_name=log_name)

  def _run(self, command):
    """Run the command."""
    print('Running: ' + command)
    return api.run(command)

  def _should_kill(self, run_bot_line):  # pylint: disable=unused-argument
    """Determine if this run_bot.py process should be killed."""
    return True

  def reboot(self):
    """Reboot the machine with `sudo reboot` and verify if succeeded."""
    try:
      api.sudo('reboot', timeout=1)
    except exceptions.CommandTimeout:
      # The timeout exception is expected if rebooting is successful.
      pass

    try:
      api.run('echo "Test rebooting"', timeout=3)
      raise Exception(
          'Failed to reboot because we can still connect to the machine.')
    except ssh_exception.ProxyCommandFailure:
      print('Cannot connect to the machine. The machine has been rebooted '
            'successfully.')

  def restart(self):
    """Restart clusterfuzz by killing existing run_bot.py processes and starting
    it back again."""
    pids = self._get_run_bot_pids()
    if not pids:
      raise Exception('No run_bot.py is running.')

    self._run('kill %s' % ' '.join(pids))
    time.sleep(3)

    for _ in range(30):
      new_pids = self._get_run_bot_pids()
      if new_pids:
        break

      time.sleep(1)

    if not new_pids:
      raise Exception('Failed to start run_bot.py after restarting.')

    print('run_bot.py has been restarted (PID=%s).' % ','.join(new_pids))

  def tail(self, log_name, line_count):
    """Print the last x lines of ./bot/logs/`log_name`.log."""
    self._run('tail -n {line_count} {log_path}'.format(
        line_count=line_count, log_path=self._log_path(log_name)))

  def tailf(self, log_names):
    """Print ./bot/logs/`name`.log in real-time (equivalent to `tail -f`)."""
    log_paths = ' '.join([self._log_path(i) for i in log_names])
    self._run('tail -f -n 100 %s' % log_paths)

  def _copy_staging_archive_from_local_to_remote(self, local_zip_path):
    """Copy staging archive from local to remote."""
    remote_zip_path = (
        '{clusterfuzz_parent_path}/{staging_source_filename}'.format(
            clusterfuzz_parent_path=self.clusterfuzz_parent_path,
            staging_source_filename=self.staging_source_filename))
    self._run('rm -f ' + remote_zip_path)
    api.put(local_zip_path, remote_zip_path)

  def stage(self, config_dir):
    """Stage a zip (built by `python butler.py package`)."""
    os.environ['CONFIG_DIR_OVERRIDE'] = config_dir

    # Restarting ensures that the target bot is updated to latest revision.
    # See crbug.com/674173 for more info.
    self.restart()

    local_zip_path = package.package(
        revision=butler_common.compute_staging_revision(),
        platform_name=self.platform)
    self._copy_staging_archive_from_local_to_remote(local_zip_path)

    self._run(('cd {clusterfuzz_parent_path} && '
               'unzip -o -d . {staging_source_filename}').format(
                   clusterfuzz_parent_path=self.clusterfuzz_parent_path,
                   staging_source_filename=self.staging_source_filename))
    self._run('chown -R {username} {clusterfuzz_parent_path}'.format(
        username=self.username,
        clusterfuzz_parent_path=self.clusterfuzz_parent_path))

    self.restart()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handlers used to recreate recurring tasks."""

from clusterfuzz._internal.base import tasks
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import ndb_utils
from clusterfuzz._internal.metrics import logs
from handlers import base_handler
from libs import handler


class OpenReproducibleTestcaseTasksScheduler(base_handler.Handler):
  """Create tasks for open reproducible testcases."""
  task = None

  @handler.cron()
  def get(self):
    """Handle a GET request."""
    assert self.task

    # Create new tasks for the open reproducible test cases.
    for status in ['Processed', 'Duplicate']:
      testcases = data_types.Testcase.query(
          ndb_utils.is_true(data_types.Testcase.open),
          ndb_utils.is_false(data_types.Testcase.one_time_crasher_flag),
          data_types.Testcase.status == status)

      for testcase in testcases:
        try:
          tasks.add_task(
              self.task,
              testcase.key.id(),
              testcase.job_type,
              queue=tasks.queue_for_testcase(testcase))
        except Exception:
          logs.log_error('Failed to add task.')
          continue


class ImpactTasksScheduler(OpenReproducibleTestcaseTasksScheduler):
  """Create impact tasks."""
  task = 'impact'


class ProgressionTasksScheduler(OpenReproducibleTestcaseTasksScheduler):
  """Create progression tasks."""
  task = 'progression'


class SimpleRecurringTaskScheduler(base_handler.Handler):
  """Recreate a recurring task."""
  task = None
  argument = 0
  job_type = 'none'

  @handler.cron()
  def get(self):
    assert self.task
    tasks.add_task(self.task, self.argument, self.job_type)


class UploadReportsTaskScheduler(SimpleRecurringTaskScheduler):
  """Recreate upload reports tasks."""
  task = 'upload_reports'



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Exception classes for minimizers."""


class MinimizationDeadlineExceededError(Exception):
  """Exception thrown if the deadline for minimization has been exceeded."""

  def __init__(self, testcase):
    Exception.__init__(self, 'Deadline exceeded.')
    self.testcase = testcase


class NoCommandError(Exception):
  """Exception thrown if no command is configured for test runs."""

  def __init__(self):
    Exception.__init__(self, 'Attempting to run with no command configured.')


class TokenizationFailureError(Exception):

  def __init__(self, minimization_type):
    Exception.__init__(self, 'Unable to perform ' + minimization_type + '.')


class AntlrDecodeError(Exception):
  """Raised when Antlr can't minimize input because it is not unicode."""



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""search_tokenizer tokenizes a string into tokens
  according to our keyword searching use cases."""

import re


def tokenize(s):
  """Tokenize a string, by line, into atomic tokens and complex tokens."""
  if not s:
    s = ''

  s = '%s' % s
  tokens = set()

  lines = s.splitlines()
  for line in lines:
    line = line.strip()
    only_ascii = re.sub(r'\s*[^\x00-\x7F]+\s*', ' ', line)
    tokens |= _complex_tokenize(only_ascii, limit=10)
    tokens.add(line.lower())

  return tokens


def tokenize_bug_information(testcase):
  """Tokenize bug information for searching."""
  bug_indices = []

  if testcase.bug_information:
    bug_indices.append(testcase.bug_information.lower().strip())
  if testcase.group_bug_information:
    bug_indices.append(str(testcase.group_bug_information))

  return bug_indices


def tokenize_impact_version(version):
  """Tokenize impact."""
  if not version:
    return []

  tokens = set()
  splitted = version.split('.')
  for index in range(len(splitted)):
    tokens.add('.'.join(splitted[0:(index + 1)]))

  return [t for t in tokens if t.strip()]


def prepare_search_keyword(s):
  """Prepare the search keywords into the form that is appropriate for searching
    according to our tokenization algorithm."""
  return s.lower().strip()


def _is_camel_case_ab(s, index):
  """Determine if the index is at 'aB', which is the start of a camel token.
    For example, with 'workAt', this function detects 'kA'."""
  return index >= 1 and s[index - 1].islower() and s[index].isupper()


def _is_camel_case_abb(s, index):
  """Determine if the index ends at 'ABb', which is the start of a camel
    token. For example, with 'HTMLParser', this function detects 'LPa'."""
  return (index >= 2 and s[index - 2].isupper() and s[index - 1].isupper() and
          s[index].islower())


def _token_indices(s):
  """Iterate through (end_current_token_index, start_next_token_index) of
    `s`, which is tokenized based on non-alphanumeric characters and camel
    casing. For example, 'aa:bbCC' have 3 tokens: 'aa', 'bb', 'CC'.
    This function iterates through (1,3), (4,5), and (6,7); they represent
    a[a]:[b]bCC, aa:b[b][C]C, and aa:bbC[C][], respectively."""
  index = 0
  length = len(s)
  while index < length:
    if not s[index].isalnum():
      end_index = index - 1
      while index < length and not s[index].isalnum():
        index += 1
      yield end_index, index
    elif _is_camel_case_ab(s, index):
      yield (index - 1), index
      index += 1
    elif _is_camel_case_abb(s, index):
      yield (index - 2), (index - 1)
      index += 1
    else:
      index += 1

  yield (length - 1), length


def _complex_tokenize(s, limit):
  """Tokenize a string into complex tokens. For example, a:b:c is tokenized into
    ['a', 'b', 'c', 'a:b', 'a:b:c', 'b:c']. This method works recursively. It
    generates all possible complex tokens starting from the first token. Then,
    it cuts off the first token and calls _complex_tokenize(..) with the rest
    of `s`.

    `limit` restricts the number of atomic tokens."""
  if not s:
    return set()

  tokens = []
  second_token_index = len(s)
  count = 0
  for end_index, next_start_index in _token_indices(s):
    tokens.append(s[0:(end_index + 1)])
    count += 1
    second_token_index = min(next_start_index, second_token_index)

    if count >= limit:
      break

  tokens = set(t.lower() for t in tokens if t.strip())
  tokens |= _complex_tokenize(s[second_token_index:], limit=limit)
  return tokens



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler that triggers the backup URL. We need this handler because
  https://cloud.google.com/appengine/articles/scheduled_backups doesn't support
  backing up all entities. Therefore, we have to get all entities here and
  send them to the backup URL."""

import datetime

from google.cloud import ndb
import googleapiclient

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.metrics import logs
from handlers import base_handler
from libs import handler

# CrashStatistic is excluded because the number of records is too high and
# can be rebuilt from BigQuery dataset.
EXCLUDED_MODELS = {'CrashStatistic', 'CrashStatisticJobHistory'}


def _datastore_client():
  """Return an api client for datastore."""
  return googleapiclient.discovery.build('datastore', 'v1')


class Handler(base_handler.Handler):
  """Handler for triggering the backup URL."""

  @handler.cron()
  def get(self):
    """Handle a cron job."""
    backup_bucket = local_config.Config(
        local_config.PROJECT_PATH).get('backup.bucket')
    if not backup_bucket:
      logs.log('No backup bucket is set, skipping.')
      return 'OK'

    kinds = [
        kind for kind in ndb.Model._kind_map  # pylint: disable=protected-access
        if (not kind.startswith('_') and kind not in EXCLUDED_MODELS)
    ]

    app_id = utils.get_application_id()
    timestamp = datetime.datetime.utcnow().strftime('%Y-%m-%d-%H:%M:%S')
    output_url_prefix = (
        'gs://{backup_bucket}/datastore-backups/{timestamp}'.format(
            backup_bucket=backup_bucket, timestamp=timestamp))
    body = {
        'output_url_prefix': output_url_prefix,
        'entity_filter': {
            'kinds': kinds
        }
    }

    try:
      request = _datastore_client().projects().export(
          projectId=app_id, body=body)
      response = request.execute()

      message = 'Datastore export succeeded.'
      status_code = 200
      logs.log(message, response=response)
    except googleapiclient.errors.HttpError as e:
      message = 'Datastore export failed.'
      status_code = e.resp.status
      logs.log_error(message, error=str(e))

    return (message, status_code, {'Content-Type': 'text/plain'})



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#===- lib/asan/scripts/asan_symbolize.py -----------------------------------===#
#
#                     The LLVM Compiler Infrastructure
#
# This file is distributed under the University of Illinois Open Source
# License. See LICENSE.TXT for details.
#
#===------------------------------------------------------------------------===#

#
# This file has local modifications needed for ClusterFuzz project.
#

# Disable all pylint warnings/errors as this is based on external code.
# pylint: disable-all

import os
import re
import subprocess
import sys

import six

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.platforms.android import fetch_artifact
from clusterfuzz._internal.platforms.android import settings
from clusterfuzz._internal.platforms.android import symbols_downloader
from clusterfuzz._internal.platforms.linux import lkl
from clusterfuzz._internal.platforms.linux.lkl import constants as lkl_constants
from clusterfuzz._internal.system import archive
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell

try:
  import pty
  import termios
except ImportError:
  # Applies only on unix platforms.
  pass

stack_inlining = 'false'
llvm_symbolizer_path = ''
pipes = []
symbolizers = {}

# 0 0x7f6e35cf2e45  (/blah/foo.so+0x11fe45)
STACK_TRACE_LINE_REGEX = re.compile(
    '^( *#([0-9]+) *)(0x[0-9a-f]+) *\(([^+]*)\+(0x[0-9a-f]+)\)')
# [    0.058680] #0 [<0x000000000062a66a>] print_address_description+0x6a/0x5c0
STACK_TRACE_LINE_REGEX_LKL = re.compile(
    r'^\[ *\d+\.\d+\]( *#([0-9]+) *)\[\<([x0-9a-fA-F]+)\>\] *'
    r'\(?(([\w]+)\+([\w]+)/[\w]+)\)?')


class LineBuffered(object):
  """Disable buffering on a file object."""

  def __init__(self, stream):
    self.stream = stream

  def write(self, data):
    self.stream.write(data)
    if '\n' in data:
      self.stream.flush()

  def __getattr__(self, attr):
    return getattr(self.stream, attr)


# Construct a path to the .dSYM bundle for the given binary.
# There are three possible cases for binary location in Chromium:
# 1. The binary is a standalone executable or dynamic library in the product
#    dir, the debug info is in "binary.dSYM" in the product dir.
# 2. The binary is a standalone framework or .app bundle, the debug info is in
#    "Framework.dSYM" or "App.dSYM" in the product dir.
# 3. The binary is a framework or an .app bundle within another .app bundle
#    (e.g. Outer.app/Contents/Versions/1.2.3.4/Inner.app), and the debug info
#    is in Inner.dSYM in the product dir.
# The first case is handled by llvm-symbolizer, so we only need to construct
# .dSYM paths for .app bundles and frameworks.
# We're assuming that there're no more than two nested bundles in the binary
# path. Only one of these bundles may be a framework and frameworks cannot
# contain other bundles.
def chrome_dsym_hints(binary):
  """Construct a path to the .dSYM bundle for the given binary.
  There are three possible cases for binary location in Chromium:
  1. The binary is a standalone executable or dynamic library in the product
     dir, the debug info is in "binary.dSYM" in the product dir.
  2. The binary is a standalone framework or .app bundle, the debug info is in
     "Framework.framework.dSYM" or "App.app.dSYM" in the product dir.
  3. The binary is a framework or an .app bundle within another .app bundle
     (e.g. Outer.app/Contents/Versions/1.2.3.4/Inner.app), and the debug info
     is in Inner.app.dSYM in the product dir.
  The first case is handled by llvm-symbolizer, so we only need to construct
  .dSYM paths for .app bundles and frameworks."""
  path_parts = binary.split(os.path.sep)
  app_positions = []
  framework_positions = []
  for index, part in enumerate(path_parts):
    if part.endswith('.app'):
      app_positions.append(index)
    elif part.endswith('.framework'):
      framework_positions.append(index)

  bundle_positions = app_positions + framework_positions
  if len(bundle_positions) == 0:
    # Case 1: this is a standalone executable or dylib.
    return []

  # Cases 2 and 3. The outermost bundle (which is the only bundle in the case 2)
  # is located in the product dir.
  bundle_positions.sort()
  outermost_bundle = bundle_positions[0]
  product_dir = path_parts[:outermost_bundle]
  # In case 2 this is the same as |outermost_bundle|.
  innermost_bundle = bundle_positions[-1]
  innermost_bundle_dir = path_parts[innermost_bundle]
  innermost_bundle_dir = utils.strip_from_right(innermost_bundle_dir, '.app')
  innermost_bundle_dir = utils.strip_from_right(innermost_bundle_dir,
                                                '.framework')
  dsym_path = product_dir + [innermost_bundle_dir]
  result = '%s.dSYM' % os.path.sep.join(dsym_path)
  return [result]


def disable_buffering():
  """Make this process and child processes stdout unbuffered."""
  os.environ['PYTHONUNBUFFERED'] = '1'

  if not isinstance(sys.stdout, LineBuffered):
    # Don't wrap sys.stdout if it is already wrapped.
    # See https://github.com/google/clusterfuzz/issues/234 for why.
    # Since sys.stdout is a C++ object, it's impossible to do sys.stdout.write =
    # lambda...
    sys.stdout = LineBuffered(sys.stdout)


def fix_filename(file_name):
  """Clean up the filename, nulls out tool specific ones."""
  file_name = re.sub('.*asan_[a-z_]*.cc:[0-9]*', '_asan_rtl_', file_name)
  file_name = re.sub('.*crtstuff.c:0', '', file_name)
  file_name = re.sub(':0$', '', file_name)

  # If we don't have a file name, just bail out.
  if not file_name or file_name.startswith('??'):
    return ''

  return os.path.normpath(file_name)


def fix_function_name(function_name):
  """Clean up function name."""
  if function_name.startswith('??'):
    return ''

  return function_name


def get_stack_frame(binary, addr, function_name, file_name):
  """Return a stack frame entry."""
  # Cleanup file and function name.
  file_name = fix_filename(file_name)
  function_name = fix_function_name(function_name)

  # Check if we don't have any symbols at all. If yes, this is probably
  # a system library. In this case, just return the binary name.
  if not function_name and not file_name:
    return '%s in %s' % (addr, os.path.basename(binary))

  # We just have a file name. Probably running in global context.
  if not function_name:
    # Filter the filename to act as a function name.
    filtered_file_name = os.path.basename(file_name)
    return '%s in %s %s' % (addr, filtered_file_name, file_name)

  # Regular stack frame.
  return '%s in %s %s' % (addr, function_name, file_name)


def is_valid_arch(s):
  """Check if this is a valid supported architecture."""
  return s in [
      "i386", "x86_64", "x86_64h", "arm", "armv6", "armv7", "armv7s", "armv7k",
      "arm64", "powerpc64", "powerpc64le", "s390x", "s390"
  ]


def guess_arch(address):
  """Guess which architecture we're running on (32/64).
  10 = len('0x') + 8 hex digits."""
  if len(address) > 10:
    return 'x86_64'
  else:
    return 'i386'


class Symbolizer(object):

  def __init__(self):
    pass

  def symbolize(self, addr, binary, offset):
    """Symbolize the given address (pair of binary and offset).

    Overridden in subclasses.
    Args:
        addr: virtual address of an instruction.
        binary: path to executable/shared object containing this instruction.
        offset: instruction offset in the @binary.
    Returns:
        list of strings (one string for each inlined frame) describing
        the code locations for this instruction (that is, function name, file
        name, line and column numbers).
    """
    return None


class LLVMSymbolizer(Symbolizer):

  def __init__(self, symbolizer_path, default_arch, system, dsym_hints=[]):
    super(LLVMSymbolizer, self).__init__()
    self.symbolizer_path = symbolizer_path
    self.default_arch = default_arch
    self.system = system
    self.dsym_hints = dsym_hints
    self.pipe = self.open_llvm_symbolizer()

  def open_llvm_symbolizer(self):
    if not os.path.exists(self.symbolizer_path):
      return None

    # Setup symbolizer command line.
    cmd = [
        self.symbolizer_path,
        '--default-arch=%s' % self.default_arch,
        '--demangle',
        '--functions=linkage',
        '--inlining=%s' % stack_inlining,
    ]
    if self.system == 'darwin':
      for hint in self.dsym_hints:
        cmd.append('--dsym-hint=%s' % hint)

    # Set LD_LIBRARY_PATH to use the right libstdc++.
    env_copy = environment.copy()
    env_copy['LD_LIBRARY_PATH'] = os.path.dirname(self.symbolizer_path)

    # FIXME: Since we are not using process_handler.run_process here, we can run
    # into issues with unicode environment variable and values. Add this
    # explicit hack to convert these into strings.
    env_copy = {str(key): str(value) for key, value in six.iteritems(env_copy)}

    # Run the symbolizer.
    pipe = subprocess.Popen(
        cmd, env=env_copy, stdin=subprocess.PIPE, stdout=subprocess.PIPE)

    global pipes
    pipes.append(pipe)

    return pipe

  def symbolize(self, addr, binary, offset):
    """Overrides Symbolizer.symbolize."""
    if not binary.strip():
      return ['%s in' % addr]

    result = []
    try:
      symbolizer_input = '"%s" %s' % (binary, offset)
      self.pipe.stdin.write(symbolizer_input.encode('utf-8') + b'\n')
      self.pipe.stdin.flush()
      while True:
        function_name = self.pipe.stdout.readline().rstrip().decode('utf-8')
        if not function_name:
          break

        file_name = self.pipe.stdout.readline().rstrip().decode('utf-8')
        result.append(get_stack_frame(binary, addr, function_name, file_name))

    except Exception:
      logs.log_error('Symbolization using llvm-symbolizer failed for: "%s".' %
                     symbolizer_input)
      result = []
    if not result:
      result = None
    return result


def LLVMSymbolizerFactory(system, default_arch, dsym_hints=[]):
  return LLVMSymbolizer(llvm_symbolizer_path, default_arch, system, dsym_hints)


class Addr2LineSymbolizer(Symbolizer):

  def __init__(self, binary):
    super(Addr2LineSymbolizer, self).__init__()
    self.binary = binary
    self.pipe = self.open_addr2line()

  def open_addr2line(self):
    cmd = ['addr2line', '--demangle', '-f', '-e', self.binary]
    pipe = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    global pipes
    pipes.append(pipe)
    return pipe

  def symbolize(self, addr, binary, offset):
    """Overrides Symbolizer.symbolize."""
    if self.binary != binary:
      return None
    if not binary.strip():
      return ['%s in' % addr]

    try:
      symbolizer_input = str(offset).encode('utf-8')
      self.pipe.stdin.write(symbolizer_input + b'\n')
      self.pipe.stdin.flush()
      function_name = self.pipe.stdout.readline().rstrip().decode('utf-8')
      file_name = self.pipe.stdout.readline().rstrip().decode('utf-8')
    except Exception:
      logs.log_error('Symbolization using addr2line failed for: "%s %s".' %
                     (binary, str(offset)))
      function_name = ''
      file_name = ''

    return [get_stack_frame(binary, addr, function_name, file_name)]


class UnbufferedLineConverter(object):
  """Wrap a child process that responds to each line of input with one line of output.

  Uses pty to trick the child into providing unbuffered output.
  """

  def __init__(self, args, close_stderr=False):
    pid, fd = pty.fork()
    if pid == 0:
      # We're the child. Transfer control to command.
      if close_stderr:
        dev_null = os.open('/dev/null', 0)
        os.dup2(dev_null, 2)
      os.execvp(args[0], args)
    else:
      # Disable echoing.
      attr = termios.tcgetattr(fd)
      attr[3] = attr[3] & ~termios.ECHO
      termios.tcsetattr(fd, termios.TCSANOW, attr)
      # Set up a file()-like interface to the child process
      self.r = os.fdopen(fd, 'r', 1)
      self.w = os.fdopen(os.dup(fd), 'w', 1)

  def convert(self, line):
    self.w.write(line + '\n')
    return self.readline()

  def readline(self):
    return self.r.readline().rstrip()


class DarwinSymbolizer(Symbolizer):

  def __init__(self, addr, binary, arch):
    super(DarwinSymbolizer, self).__init__()
    self.binary = binary
    self.arch = arch
    self.open_atos()

  def open_atos(self):
    cmdline = ['atos', '-o', self.binary, '-arch', self.arch]
    self.atos = UnbufferedLineConverter(cmdline, close_stderr=True)

  def symbolize(self, addr, binary, offset):
    """Overrides Symbolizer.symbolize."""
    if self.binary != binary:
      return None

    try:
      atos_line = self.atos.convert('0x%x' % int(offset, 16))
      while 'got symbolicator for' in atos_line:
        atos_line = self.atos.readline()
      # A well-formed atos response looks like this:
      #   foo(type1, type2) (in object.name) (filename.cc:80)
      match = re.match('^(.*) \(in (.*)\) \((.*:\d*)\)$', atos_line)
      if match:
        function_name = match.group(1)
        function_name = re.sub('\(.*?\)', '', function_name)
        file_name = match.group(3)
        return [get_stack_frame(binary, addr, function_name, file_name)]
      else:
        return ['%s in %s' % (addr, atos_line)]
    except Exception:
      logs.log_error('Symbolization using atos failed for: "%s %s".' %
                     (binary, str(offset)))
      return ['{} ({}:{}+{})'.format(addr, binary, self.arch, offset)]


# Chain several symbolizers so that if one symbolizer fails, we fall back
# to the next symbolizer in chain.
class ChainSymbolizer(Symbolizer):

  def __init__(self, symbolizer_list):
    super(ChainSymbolizer, self).__init__()
    self.symbolizer_list = symbolizer_list

  def symbolize(self, addr, binary, offset):
    """Overrides Symbolizer.symbolize."""
    for symbolizer in self.symbolizer_list:
      if symbolizer:
        result = symbolizer.symbolize(addr, binary, offset)
        if result:
          return result
    return None

  def append_symbolizer(self, symbolizer):
    self.symbolizer_list.append(symbolizer)


def SystemSymbolizerFactory(system, addr, binary, arch):
  if system == 'darwin':
    return DarwinSymbolizer(addr, binary, arch)
  elif system.startswith('linux'):
    return Addr2LineSymbolizer(binary)


class SymbolizationLoop(object):

  def __init__(self, binary_path_filter=None, dsym_hint_producer=None):
    # Used by clients who may want to supply a different binary name.
    # E.g. in Chrome several binaries may share a single .dSYM.
    self.binary_path_filter = binary_path_filter
    self.dsym_hint_producer = dsym_hint_producer
    self.system = sys.platform
    self.llvm_symbolizers = {}
    self.last_llvm_symbolizer = None
    self.dsym_hints = set([])
    self.lkl_binary_name = None

  def symbolize_address(self, addr, binary, offset, arch):
    # On non-Darwin (i.e. on platforms without .dSYM debug info) always use
    # a single symbolizer binary.
    # On Darwin, if the dsym hint producer is present:
    #  1. check whether we've seen this binary already; if so,
    #     use |llvm_symbolizers[binary]|, which has already loaded the debug
    #     info for this binary (might not be the case for
    #     |last_llvm_symbolizer|);
    #  2. otherwise check if we've seen all the hints for this binary already;
    #     if so, reuse |last_llvm_symbolizer| which has the full set of hints;
    #  3. otherwise create a new symbolizer and pass all currently known
    #     .dSYM hints to it.
    if not binary in self.llvm_symbolizers:
      use_new_symbolizer = True
      if self.system == 'darwin' and self.dsym_hint_producer:
        dsym_hints_for_binary = set(self.dsym_hint_producer(binary))
        use_new_symbolizer = bool(dsym_hints_for_binary - self.dsym_hints)
        self.dsym_hints |= dsym_hints_for_binary
      if self.last_llvm_symbolizer and not use_new_symbolizer:
        self.llvm_symbolizers[binary] = self.last_llvm_symbolizer
      else:
        self.last_llvm_symbolizer = LLVMSymbolizerFactory(
            self.system, arch, self.dsym_hints)
        self.llvm_symbolizers[binary] = self.last_llvm_symbolizer

    # Use the chain of symbolizers:
    # LLVM symbolizer -> addr2line/atos
    # (fall back to next symbolizer if the previous one fails).
    if not binary in symbolizers:
      symbolizers[binary] = ChainSymbolizer([self.llvm_symbolizers[binary]])
    result = symbolizers[binary].symbolize(addr, binary, offset)
    if result is None:
      # Initialize system symbolizer only if other symbolizers failed.
      symbolizers[binary].append_symbolizer(
          SystemSymbolizerFactory(self.system, addr, binary, arch))
      result = symbolizers[binary].symbolize(addr, binary, offset)
    # The system symbolizer must produce some result.
    assert result
    return result

  def _line_parser(self, line):
    """Parses line for frameno_str, addr, binary, offset, arch."""
    match = STACK_TRACE_LINE_REGEX.match(line)
    if match:
      _, frameno_str, addr, binary, offset = match.groups()
      arch = ""
      # Arch can be embedded in the filename, e.g.: "libabc.dylib:x86_64h"
      colon_pos = binary.rfind(":")
      if colon_pos != -1:
        maybe_arch = binary[colon_pos + 1:]
        if is_valid_arch(maybe_arch):
          arch = maybe_arch
          binary = binary[0:colon_pos]
      if arch == "":
        arch = guess_arch(addr)

      return frameno_str, addr, binary, offset, arch

    return None, None, None, None, None

  def _lkl_line_parser(self, line):
    """Parses line for frameno_str, addr, binary, offset, arch."""
    match = STACK_TRACE_LINE_REGEX_LKL.match(line)
    if match:
      # LKL is always ran on x86_64 host.
      arch = 'x86_64'
      _, frameno_str, addr, _, _, _ = match.groups()
      # In LKL the offset from the base of kernel lib is the address.
      offset = addr
      binary = self.lkl_binary_name
      return frameno_str, addr, binary, offset, arch

    return None, None, None, None, None

  def process_stacktrace(self, unsymbolized_crash_stacktrace):
    self.frame_no = 0
    symbolized_crash_stacktrace = u''
    unsymbolized_crash_stacktrace_lines = \
      unsymbolized_crash_stacktrace.splitlines()
    if lkl.is_lkl_stack_trace(unsymbolized_crash_stacktrace):
      line_parser = self._lkl_line_parser
      self.lkl_binary_name = lkl.get_lkl_binary_name(
          unsymbolized_crash_stacktrace_lines)
      # This should never happen but if it does, lets just return the unsymbolized stack.
      # We can't symbolize anything anyways.
      if not self.lkl_binary_name:
        return unsymbolized_crash_stacktrace
    else:
      line_parser = self._line_parser

    for line in unsymbolized_crash_stacktrace_lines:
      self.current_line = utils.decode_to_unicode(line.rstrip())
      frameno_str, addr, binary, offset, arch = line_parser(line)
      if not binary or not offset:
        symbolized_crash_stacktrace += u'%s\n' % self.current_line
        continue

      if frameno_str == '0':
        # Assume that frame #0 is the first frame of new stack trace.
        self.frame_no = 0
      original_binary = binary
      if self.binary_path_filter:
        binary = self.binary_path_filter(binary)
      symbolized_line = self.symbolize_address(addr, binary, offset, arch)
      if not symbolized_line:
        if original_binary != binary:
          symbolized_line = self.symbolize_address(addr, original_binary,
                                                   offset, arch)

      if not symbolized_line:
        symbolized_crash_stacktrace += u'%s\n' % self.current_line
      else:
        for symbolized_frame in symbolized_line:
          symbolized_crash_stacktrace += u'%s\n' % (
              '    #' + str(self.frame_no) + ' ' + symbolized_frame.rstrip())
          self.frame_no += 1

    # Close any left-over open pipes.
    for pipe in pipes:
      pipe.stdin.close()
      pipe.stdout.close()
      pipe.kill()

    return symbolized_crash_stacktrace


def filter_binary_path(binary_path):
  """Filters binary path to provide a local copy."""
  if environment.is_android() or environment.is_lkl_job():
    return symbols_downloader.filter_binary_path(binary_path)

  if environment.platform() == 'CHROMEOS':
    # FIXME: Add code to pull binaries from ChromeOS device.
    return binary_path

  if environment.is_chromeos_system_job():
    # This conditional is True for ChromeOS system fuzzers that are running on
    # Linux. Ensure that the binary is always looked for in the chroot and not
    # in system directories.
    build_dir = environment.get_value('BUILD_DIR')
    if not binary_path.startswith(build_dir):
      # Fixup path so |binary_path| points to a binary in the chroot (probably
      # a system library).
      return os.path.join(build_dir, binary_path[1:])

  # For Linux and Mac, the binary exists locally. No work to do,
  # just return the same binary path.
  return binary_path


def symbolize_stacktrace(unsymbolized_crash_stacktrace,
                         enable_inline_frames=True):
  """Symbolize a crash stacktrace."""
  if environment.is_trusted_host():
    from clusterfuzz._internal.bot.untrusted_runner import symbolize_host
    return symbolize_host.symbolize_stacktrace(unsymbolized_crash_stacktrace,
                                               enable_inline_frames)

  platform = environment.platform()
  if platform == 'WINDOWS':
    # Windows Clang ASAN provides symbolized stacktraces anyway.
    return unsymbolized_crash_stacktrace

  if platform == 'FUCHSIA':
    # Fuchsia Clang ASAN provides symbolized stacktraces anyway.
    return unsymbolized_crash_stacktrace

  # FIXME: Support symbolization on ChromeOS device.
  if platform == 'CHROMEOS':
    return unsymbolized_crash_stacktrace

  # Initialize variables.
  global llvm_symbolizer_path
  global pipes
  global stack_inlining
  global symbolizers
  pipes = []
  stack_inlining = str(enable_inline_frames).lower()
  symbolizers = {}

  # Make sure we have a llvm symbolizer for this platform.
  llvm_symbolizer_path = environment.get_llvm_symbolizer_path()
  if not llvm_symbolizer_path:
    return unsymbolized_crash_stacktrace

  # Disable buffering for stdout.
  disable_buffering()

  loop = SymbolizationLoop(
      binary_path_filter=filter_binary_path,
      dsym_hint_producer=chrome_dsym_hints)
  symbolized_crash_stacktrace = loop.process_stacktrace(
      unsymbolized_crash_stacktrace)

  return symbolized_crash_stacktrace



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tasks RPC implementations."""

from google.protobuf import wrappers_pb2
from google.protobuf.any_pb2 import Any
import six

from clusterfuzz._internal.bot import testcase_manager
from clusterfuzz._internal.bot.tasks import corpus_pruning_task
from clusterfuzz._internal.bot.tasks import fuzz_task
from clusterfuzz._internal.bot.tasks import minimize_task
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.protos import untrusted_runner_pb2
from clusterfuzz.fuzz import engine


def _proto_to_fuzz_target(proto):
  """Convert protobuf to FuzzTarget."""
  return data_types.FuzzTarget(
      engine=proto.engine, project=proto.project, binary=proto.binary)


def _proto_to_cross_pollinate_fuzzer(proto):
  """Convert protobuf to CrossPollinateFuzzer."""
  return corpus_pruning_task.CrossPollinateFuzzer(
      fuzz_target=_proto_to_fuzz_target(proto.fuzz_target),
      backup_bucket_name=proto.backup_bucket_name,
      corpus_engine_name=proto.corpus_engine_name)


def prune_corpus(request, _):
  """Prune corpus."""
  context = corpus_pruning_task.Context(
      _proto_to_fuzz_target(request.fuzz_target), [
          _proto_to_cross_pollinate_fuzzer(proto)
          for proto in request.cross_pollinate_fuzzers
      ])

  result = corpus_pruning_task.do_corpus_pruning(
      context, request.last_execution_failed, request.revision)

  cross_pollination_stats = None
  if result.cross_pollination_stats:
    cross_pollination_stats = untrusted_runner_pb2.CrossPollinationStats(
        project_qualified_name=result.cross_pollination_stats.
        project_qualified_name,
        method=result.cross_pollination_stats.method,
        sources=result.cross_pollination_stats.sources,
        tags=result.cross_pollination_stats.tags,
        initial_corpus_size=result.cross_pollination_stats.initial_corpus_size,
        corpus_size=result.cross_pollination_stats.corpus_size,
        initial_edge_coverage=result.cross_pollination_stats.
        initial_edge_coverage,
        edge_coverage=result.cross_pollination_stats.edge_coverage,
        initial_feature_coverage=result.cross_pollination_stats.
        initial_feature_coverage,
        feature_coverage=result.cross_pollination_stats.feature_coverage)

  # Intentionally skip edge and function coverage values as those would come
  # from fuzzer coverage cron task (see src/go/server/cron/coverage.go).
  coverage_info = untrusted_runner_pb2.CoverageInfo(
      corpus_size_units=result.coverage_info.corpus_size_units,
      corpus_size_bytes=result.coverage_info.corpus_size_bytes,
      corpus_location=result.coverage_info.corpus_location,
      corpus_backup_location=result.coverage_info.corpus_backup_location,
      quarantine_size_units=result.coverage_info.quarantine_size_units,
      quarantine_size_bytes=result.coverage_info.quarantine_size_bytes,
      quarantine_location=result.coverage_info.quarantine_location)

  crashes = [
      untrusted_runner_pb2.CorpusCrash(
          crash_state=crash.crash_state,
          crash_type=crash.crash_type,
          crash_address=crash.crash_address,
          crash_stacktrace=crash.crash_stacktrace,
          unit_path=crash.unit_path,
          security_flag=crash.security_flag,
      ) for crash in result.crashes
  ]

  return untrusted_runner_pb2.PruneCorpusResponse(
      coverage_info=coverage_info,
      crashes=crashes,
      fuzzer_binary_name=result.fuzzer_binary_name,
      revision=result.revision,
      cross_pollination_stats=cross_pollination_stats)


def process_testcase(request, _):
  """Process testcase."""
  tool_name_map = {
      untrusted_runner_pb2.ProcessTestcaseRequest.MINIMIZE: 'minimize',
      untrusted_runner_pb2.ProcessTestcaseRequest.CLEANSE: 'cleanse',
  }

  # TODO(ochang): Support other engines.
  assert request.engine == 'libFuzzer'
  assert request.operation in tool_name_map

  result = minimize_task.run_libfuzzer_engine(
      tool_name_map[request.operation], request.target_name, request.arguments,
      request.testcase_path, request.output_path, request.timeout)

  return untrusted_runner_pb2.EngineReproduceResult(
      return_code=result.return_code,
      time_executed=result.time_executed,
      output=result.output)


def _pack_values(values):
  """Pack protobuf values."""
  packed = {}
  if values is None:
    return packed

  for key, value in six.iteritems(values):
    packed_value = Any()
    if isinstance(value, float):
      packed_value.Pack(wrappers_pb2.DoubleValue(value=value))
    elif isinstance(value, six.integer_types):
      packed_value.Pack(wrappers_pb2.Int64Value(value=value))
    elif isinstance(value, six.string_types):
      packed_value.Pack(wrappers_pb2.StringValue(value=value))
    else:
      raise ValueError('Unknown stat type for ' + key)

    packed[key] = packed_value

  return packed


def engine_fuzz(request, _):
  """Run engine fuzzer."""
  engine_impl = engine.get(request.engine)
  result, fuzzer_metadata, strategies = fuzz_task.run_engine_fuzzer(
      engine_impl, request.target_name, request.sync_corpus_directory,
      request.testcase_directory)

  crashes = [
      untrusted_runner_pb2.EngineCrash(
          input_path=crash.input_path,
          stacktrace=crash.stacktrace,
          reproduce_args=crash.reproduce_args,
          crash_time=crash.crash_time) for crash in result.crashes
  ]

  packed_stats = _pack_values(result.stats)
  packed_strategies = _pack_values(strategies)

  return untrusted_runner_pb2.EngineFuzzResponse(
      logs=result.logs,
      command=result.command,
      crashes=crashes,
      stats=packed_stats,
      time_executed=result.time_executed,
      fuzzer_metadata=fuzzer_metadata,
      strategies=packed_strategies)


def engine_reproduce(request, _):
  """Run engine reproduce."""
  engine_impl = engine.get(request.engine)
  result = testcase_manager.engine_reproduce(engine_impl, request.target_name,
                                             request.testcase_path,
                                             request.arguments, request.timeout)
  return untrusted_runner_pb2.EngineReproduceResult(
      command=result.command,
      return_code=result.return_code,
      time_executed=result.time_executed,
      output=result.output)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Build attributes of every testcase."""

import datetime
import sys

import six

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import ndb_utils
from local.butler.scripts import attribute_builder
from local.butler.scripts import batcher

BATCH_SIZE = 500


def to_dict(entity):
  """Convert a db.Model instance to a dict."""
  entity_dict = entity.to_dict()
  entity_dict['id'] = entity.key.id()

  for k, v in six.iteritems(entity_dict):
    if isinstance(v, datetime.datetime):
      entity_dict[k] = utils.utc_datetime_to_timestamp(v)

  return entity_dict


def get_diff(before, after):
  """Return differences in string between the two dicts, before and after."""
  diffs = []
  for k, v in six.iteritems(before):
    if k in after:
      if v != after[k]:
        diffs.append((k, (v, after[k])))
    else:
      diffs.append((k, (v, '<MISSING>')))

  for k, v in six.iteritems(after):
    if k not in before:
      diffs.append((k, ('<MISSING>', v)))

  diffs.sort()

  s = ''
  for (key, (before_value, after_value)) in diffs:
    s += '%s:\n' % key
    s += '-%s\n' % before_value
    s += '+%s\n\n' % after_value

  return s


def execute(args):
  """Build keywords."""
  count_diff = 0

  query = data_types.Testcase.query().order(-data_types.Testcase.timestamp)
  for testcases in batcher.iterate(query, BATCH_SIZE):
    for testcase in testcases:
      before_testcase = to_dict(testcase)
      attribute_builder.populate(testcase)
      after_testcase = to_dict(testcase)

      diff = get_diff(before_testcase, after_testcase)
      if (count_diff % 10) == 0 and diff:
        print('Migrate (dry=%s) id:%s\n%s' % (not args.non_dry_run,
                                              testcase.key.id(), diff))

      if diff:
        count_diff += 1

    if args.non_dry_run:
      try:
        ndb_utils.put_multi(testcases)
      except Exception:
        for testcase in testcases:
          try:
            testcase.put()
          except Exception:
            print('Error: %s %s' % (testcase.key.id(), sys.exc_info()))

  print('Done (count_diff=%d)' % count_diff)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Builtin fuzzers."""

from clusterfuzz._internal.bot.fuzzers.afl import fuzzer as afl
from clusterfuzz._internal.bot.fuzzers.libFuzzer import fuzzer as libFuzzer

BUILTIN_FUZZERS = {
    'afl': afl.Afl(),
    'libFuzzer': libFuzzer.LibFuzzer(),
}


def get(fuzzer_name):
  """Get the builtin fuzzer with the given name, or None."""
  if fuzzer_name not in BUILTIN_FUZZERS:
    return None

  return BUILTIN_FUZZERS[fuzzer_name]
